
:- ensure_loaded( '../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( opts_select_infiles/3 ).
:- requires( file_to_list_of_lines/2 ).

diff( Opts ) :-
	opts_have_help( Opts ),
	!,
	write( 'diff:              report line differences in two files.' ),
	nl, nl,
	write( 'Options.' ), nl,
	write( '  file(F)          for the files. Atoms are also taken to files' ),
	nl.    

diff( Opts ) :-
	opts_select_infiles( Opts, Fs, _Oth ),
	( Fs = [Fst,Sec|T] -> 
		( T == [] ->
			true
			;
			write( user_error, more_than_two_files_present-ignoring(T) ),
			nl( user_error )
		),
		diff_two( Fst, Sec )
		;
		write( user_error, at_least_two_files_should_be_given(Opts) ),
		nl( user_error )
	).

diff_two( Fst, Sec ) :-
	file_to_list_of_lines( Fst, FstLns ),
	file_to_list_of_lines( Sec, SecLns ),
	diff_two_lines( FstLns, SecLns, 1, 0 ).

diff_two_lines( [], SecLns, _Ln, Diff ) :-
	( SecLns == [] -> 
		ExLnsLength is 0
		; 
		length( SecLns, ExLnsLength ),
		write( there_are_extra_lines_at_end_of_second_file(ExLnsLength) ),
		nl, nl
	),
	Total is Diff + ExLnsLength,
	write( total_diff_lines(Total) ), nl.

diff_two_lines( [Hfst|Tfst], SecLns, Ln, InDiff ) :-
	( SecLns = [Hsec|Tsec] ->
		RstF = Tfst, RstS = Tsec,
		( Hfst == Hsec ->
			NxtDiff is InDiff
			;
			atom_codes( FstAtm, Hfst ),
			atom_codes( SecAtm, Hsec ),
			NxtDiff is InDiff + 1,
			write( Ln-(f(FstAtm),s(SecAtm)) ), nl
		)
		;
		length( [Hfst|Tfst], ExLnsLength ),
		write( there_are_extra_lines_at_end_of_first_file(ExLnsLength) ),
		nl, nl,
		RstF = [],
		RstS = [],
		NxtDiff is InDiff + ExLnsLength
	),
	NxLn is Ln + 1,
	diff_two_lines( RstF, RstS, NxLn, NxtDiff ).
