
:- ensure_loaded( '../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( opts_select_infiles/3 ).
:- requires( file_to_list_of_lines/2 ).

% upsh ~/islp/mcmcms/auxil/diff_ratio *ratio

diff_ratio( Opts ) :-
	opts_have_help( Opts ),
	!,
	write( 'diff_ratio:        report line differences in two ratio files.' ),
	nl, nl,
	write( 'Options.' ), nl,
	write( '  file(F)          for the files. Atoms are also taken to files' ),
	nl.    

diff_ratio( Opts ) :-
	opts_select_infiles( Opts, Fs, _Oth ),
	( Fs = [Fst,Sec|T] -> 
		( T == [] ->
			true
			;
			write( user_error, more_than_two_files_present-ignoring(T) ),
			nl( user_error )
		),
		diff_two_ratio( Fst, Sec )
		;
		write( user_error, at_least_two_files_should_be_given(Opts) ),
		nl( user_error )
	).

diff_two_ratio( Fst, Sec ) :-
	file_to_list_of_lines( Fst, FstLns ),
	file_to_list_of_lines( Sec, SecLns ),
	diff_two_ratio_lines( FstLns, SecLns, 0 ).

diff_two_ratio_lines( [], SecLns, KL ) :-
	( SecLns == [] -> 
		true
		; 
		length( SecLns, ExLnsLength ),
		write( user_error, there_are_extra_lines_at_end_of_second_file(ExLnsLength) ),
		nl( user_error ), nl( user_error )
	),
	write( kl(KL) ), nl.

diff_two_ratio_lines( [Hfst|Tfst], SecLns, KLAcc ) :-
	( SecLns = [Hsec|Tsec] ->
		append( Hfst, ".", HfstPrd ),
		append( Hsec, ".", HsecPrd ),
		read_term_from_chars( HfstPrd, TrmF, [] ),
		read_term_from_chars( HsecPrd, TrmS, [] ),
		Diff is (TrmF) - (TrmS),
		write( Diff ), nl,
		RstF = Tfst,
		NxKLAcc is KLAcc + (TrmF * log((TrmF/TrmS)) )
		;
		length( [Hfst|Tfst], ExLnsLength ),
		write( user_error, there_are_extra_lines_at_end_of_first_file(ExLnsLength) ),
		nl( user_error ), nl( user_error ),
		RstF = [],
		Tsec = [],
		NxKLAcc is KLAcc
	),
	diff_two_ratio_lines( RstF, Tsec, NxKLAcc ).
