:- ensure_loaded( '../src/init_lib' ).
:- ensure_loaded( library(lists) ).				% is_list/1, select/3.
:- ensure_loaded( library(mktemp_in_dir) ).			% /3.
:- ensure_loaded( library(flatoms) ).				% /2.
:- ensure_loaded( library(gnucomfile) ).			% /3.
:- ensure_loaded( library(file_to_list_of_lines) ).	% /2.
:- ensure_loaded( library(max_min_list) ).			% /3.
:- ensure_loaded( library(lines_to_first_numbers) ).	% /3.
:- ensure_loaded( library(pad_ranges) ).			% /5.
:- ensure_loaded( library(sys_info) ).				% /1.
:- ensure_loaded( library(gzip_files) ).			% /1.
:- ensure_loaded( library(select_compound_w_gunzip) ).	% /4.
:- ensure_loaded( library(select_all) ).	          % /4.

% upsh gsplay multi=true pad=both='%'=10 ypad=min=abs=0 nd7_idx_tms_srt_dspt.txt nd7_noidx_tms_srt_dspt.txt
% upsh gsplay multi=true pad=both='%'=10 ypad=min=abs=0 ypad=max=abs=5 nd7_idx_tms_srt_dspt.txt nd7_noidx_tms_srt_dspt.txt output=ofoul.ps term='postscript color'
% To include the above ps file, use graphicx package and something like
% \includegraphics*[width=0.8\textwidth,angle=-90]{malakako}
% 
% example from displaying .visits as boxes:
% upsh gsplay tr_cd0_fm__null_i10K__s1.visits lines=boxes ypad=min=abs=0 ypad=max=abs=7000
% gsplay( [lines(Bool),multi(Bool),...,GnuOpts,...|Files] ).
% 
% 2004/10/20, upsh gsplay cm_tas_d10K_mrm_buc_r10M_ppc_hbd_s1.lne_e1 xpad=min=0 xpad=max=10000010 ypad=min=0 ypad=max=1
% 


gsplay( In ) :-
	(In == help; In == h; memberchk(h,In); memberchk(help,In)),
	!, nl,
	write( 'gsplay:    provides a prolog interface to gnuploting files of numbers.' ), nl,
	write( '  Options.' ), nl,
	write( '        lines/1    takes boolean, boxes or uboxes; whether to use lines in plot,' ), nl,
	write( '                   def=true' ), nl,
	write( '        multi/1    boolean; wheter to plot all input files to one plot,' ), nl,
	write( '                   def=false' ), nl,
	write( '        output/1   file for output plot' ), nl,
	write( '        term/1     kind of terminal to use, e.g. \'postscript color\'' ), nl, 
	write( '        pad/1      padding rules to manipulate plot ranges '), nl,
	write( '                   e.g. ypad=min=abs=0 ypad=max=abs=5' ), nl,
	write( '        rm/1       boolean; whether to remove gnuplot commands file, def=true' ), nl,
	write( '        title/1    tile to use. It defaults to concat of the plotted files' ), nl, nl,
	write( '        All non compound options are taken to be files to plot.' ),
	nl,
	write( '        Each file should contain numbers as expected by gnuplot.' ), nl, nl.


gsplay( In ) :-
	( is_list(In) ->
		( select(lines(Lines),In,LnArgs) ->
			true
			;
			Lines = true,
			LnArgs = In
		)
		;
		Lines = true,
		LnArgs = [In]
	),
	select_pad_options( LnArgs, Pad, PadArgs ),
	% write( pad(Pad) ), nl,
	( select(rm(Rm),PadArgs,RmArgs) -> true ; Rm=true, RmArgs=PadArgs ),
	mktemp_in_dir( gnuplotcoms_XXXXXX, '.', GnuComsFile ),
	( select(multi(Multi),RmArgs,MulArgs) ->
		true ; Multi= false, MulArgs = RmArgs ),
     select_all( MulArgs, set(_), SetOpts, NonSetArgs ),
	( select(key(KOnOff),NonSetArgs,KeyArgs) -> 
		atom_concat( 'key ', KOnOff, SetK ),
		KeyOpts = [set(SetK)|SetOpts] % what happens to title here ?
		;
		SetK = 'key off',

		KeyOpts = [set(SetK),set(Title)|SetOpts],
		KeyArgs = NonSetArgs
	),
     ( select(title(InTitle),KeyArgs,TtlArgs) -> 
          files_to_title( [InTitle], Title )
          ;
          Title = FilesTitle,
          TtlArgs = KeyArgs
     ),
	select_compound_w_gunzip( TtlArgs, SelGnuOpts, Files, Gz ),
	files_to_title( Files, FilesTitle ),
	append( KeyOpts, SelGnuOpts, GnuOpts ),
	( Multi == true ->
		gsplay_2( Files, Lines, Pad, Rm, GnuOpts, GnuComsFile )
		;
		gsplay_1( Files, Lines, Pad, Rm, GnuOpts, GnuComsFile )
	),
	gzip_files( Gz ).

gsplay_1( [], _, _Pad, _Rm, _GnuOpts, _ ).
gsplay_1( [H|T], Lines, Pad, Rm, GnuOpts, GnuComsFile ) :-
	( Rm == false -> Psfx = '\"' ; atom_concat( '\"; rm -f ', GnuComsFile, Psfx ) ),
	flatoms( ['gnuplot -persist \"',GnuComsFile,Psfx], Gplot1 ),
	( Lines == false ->
		Opts1 = GnuOpts ;
		( Lines = true -> 
			Opts1 = [with(lines)|GnuOpts]
			;
			Opts1 = [with(Lines)|GnuOpts] 
		)
	),
	% ( 
	%	gnucomfile( GnuComsFile, H )
	%	;
	%	% gnucomfile_w_lines( GnuComsFile, H )
	%),
	( conclusive_padding( Pad, _, _, XPadMin/XPadMax, YPadMin/YPadMax ) ->
		true
		;
		synchronise_ranges( [H], Xmin/Xmax, Ymin/Ymax ),
		pad_ranges( Pad, Xmin/Xmax, Ymin/Ymax, XPadMin/XPadMax, YPadMin/YPadMax )
	),
	Opts = [x(XPadMin,XPadMax),y(YPadMin,YPadMax)|Opts1],
	gnucomfile( GnuComsFile, Opts, H ),
	sys_info( Gplot1 ),
	gsplay_1( T, Lines, _Pad, Rm, GnuOpts, GnuComsFile ).
	% add use of Pad. if not empty find the max and min values, etc...

gsplay_2( Files, Lines, Pad, Rm, GnuOpts, GnuComsFile ) :-
	( Rm == false ->  Psfx = '\"'; atom_concat( '\"; rm -f ', GnuComsFile, Psfx )  ),
	flatoms( ['gnuplot -persist \"',GnuComsFile,Psfx], Gplot2 ),
	( Lines == false -> Opts1 = GnuOpts ; Opts1 = [with(lines)|GnuOpts] ),
	( conclusive_padding( Pad, _, _, XPadMin/XPadMax, YPadMin/YPadMax ) ->
		true
		;
		% input option for padding
		synchronise_ranges( Files, Xmin/Xmax, Ymin/Ymax ),
		pad_ranges( Pad, Xmin/Xmax, Ymin/Ymax, XPadMin/XPadMax, YPadMin/YPadMax )	
	),
	Opts = [x((XPadMin,XPadMax)),y((YPadMin,YPadMax))|Opts1],
	gnucomfile( GnuComsFile, Opts, Files ),
	sys_info( Gplot2 ).

synchronise_ranges( [H|T], X, Y ) :-
	file_has_gnuplot_ranges( H, Xh, Yh ),
	synchronise_ranges_1( T, Xh, Yh, X, Y ).

synchronise_ranges_1( [], X, Y, X, Y ).
synchronise_ranges_1( [H|T], Xs, Ys, X, Y ) :-
	file_has_gnuplot_ranges( H, Xh, Yh ),
	update_gnuplot_range( Xs, Xh, Xn ),
	update_gnuplot_range( Ys, Yh, Yn ),
	synchronise_ranges_1( T, Xn, Yn, X, Y ).

update_gnuplot_range( CurMin/CurMax, ThisMin/ThisMax, OutMin/OutMax ) :-
	( ThisMax > CurMax -> OutMax = ThisMax ; OutMax = CurMax ),
	( ThisMin < CurMin -> OutMin = ThisMin ; OutMin = CurMin ).

file_has_gnuplot_ranges( H, X, Y ) :-
	file_to_list_of_lines( H, Lines ),
	length( Lines, Length ),
	Lminus is Length - 1,
	X = 0/Lminus,
	lines_to_first_numbers( Lines, Numbers ),
	max_min_list( Numbers, YMax, YMin ),
	Y = YMin/YMax.

% select_pad_options( List, Paddings, Rem ) :-
	% ( select(pad(Pad),List,List1) -> P1 = [pad(Pad)] ; P1 = [],List1 = List ),
	% ( select(xpad(XPad),List1,List2) -> P2 = [xpad(XPad)] ; P2 = [],List2 = List1 ),
	% % ( select(ypad(YPad),List2,Rem) -> P3 = [ypad(YPad)] ; P3 = [],Rem = List2 ),
	% append( P2, P3, P23 ),
	% append( P1, P23, Paddings ).

select_pad_options( [], [], [] ).
select_pad_options( [H|T], Paddings, Rem ) :-
	( compound( H ) ->
		H =.. [Name|_Args],
		atom_codes( Name, NmCodes ),
		( (NmCodes="pad";NmCodes=[_|"pad"]) ->
			Paddings = [H|TPaddings],
			Rem = TRem
			;
			Paddings = TPaddings,
			Rem = [H|TRem]
		)
		;
		Paddings = TPaddings,
		Rem = [H|TRem]
	),
	select_pad_options( T, TPaddings, TRem ).

conclusive_padding( [], XR, YR, XR, YR ) :-
	!,
	ground( XR ), ground( YR ).
conclusive_padding( [xpad(XRule)|T], AccX, AccY, XR, YR ) :-
	!, 
	conclusive_rule( XRule, AccX, NxtX ),
	conclusive_padding( T, NxtX, AccY, XR, YR ).
conclusive_padding( [ypad(YRule)|T], AccX, AccY, XR, YR ) :-
	!, 
	conclusive_rule( YRule, AccY, NxtY ),
	conclusive_padding( T, AccX, NxtY, XR, YR ).

% this clause is probably not so useful...
% conclusive_rule( both(What), _IMin/_IMax, Min/Max ) :-
	% conclusive_end_point(What,Min),
	% conclusive_end_point(What,Max).
conclusive_rule( max(What), IMin/_IMax, IMin/Max ) :-
	conclusive_end_point(What,Max).
conclusive_rule( min(What), _IMin/IMax, Min/IMax ) :-
	conclusive_end_point(What,Min).

conclusive_end_point( NorAbs, Val ) :-
	( number(NorAbs) -> Val is NorAbs 
		; 
		NorAbs = abs(Val)
	).

files_to_title( [File], FileQ ) :-
	!,
	atom_concat( File, '"', FileLQ ),
	atom_concat( 'title "', FileLQ, FileQ ).
files_to_title( _Files, 'title "multifile titles not yet implemented"' ).
