
:- ensure_loaded( '../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( options_load_files/1 ).
:- requires( options_multi_lists_cohese/3 ).
:- requires( select_compound_w_gunzip/4 ).
:- requires( options_gzip_at_end/3 ).
:- requires( write_terms/4 ).

% bns/test: 
% upsh ../../auxil/min_exp_loss_naive.pl load=atm=../auxil/rel_loss tr_cp28_fm_jc_asia_data_all_un_2p_i1K__s1.mvisits loss=loss_01

min_exp_loss_naive( Opts ) :-
	opts_have_help( Opts ),
	!, nl,
	write( 'min_exp_loss_naive         compute the minimum expected loss model, ' ), nl, 
	write( '                           of a chain. This is the naive version.' ), nl,
	write( '                           All models visited are considered, and' ), nl,
	write( '                           all espected losses are recorded.' ), nl,
	write( '                           E(L) = Sum [ p(M|D) L(M,^M ]' ), nl,
	write( '                           M is the model, L is the loss funtion' ), nl, 
	write( '                           and ^M is the model that argmin(s) the' ), nl,
	write( '                           summation.' ), nl,
	write( '  Options ' ), nl,
	write( '     load/1,2              files to load, usually contain loss funtion(s) ' ), nl,
	write( '     loss/1                the loss function(s) to use. Multiple and lists' ), nl,
	write( '                           are allowed.' ), nl,
	write( '     gz/1                  Boolean, gzip at end or not. Default is to' ), nl,
	write( '                           gzip any files that were gunzipped by this' ), nl,
	write( '                           script.' ), nl,
	nl, write( '     All non compound arguments are taken to be mvisits files. for each,' ), nl,
	write( '     the minimum expected loss for given loss function will be computed.' ), nl, nl.

min_exp_loss_naive( Opts ) :-
	options_load_files( Opts ),
	options_multi_lists_cohese( Opts, loss, LFns ),
	select_compound_w_gunzip( Opts, _, Files, Gz ),
	min_exp_loss_naive_fs_ls( Files, LFns ),
	options_gzip_at_end( Opts, Files, Gz ).

min_exp_loss_naive_fs_ls( [], _LFns ).
min_exp_loss_naive_fs_ls( [Fl|Fls], LFns ) :-
	mvisits_file_to_frq_list( Fl, NofVs, Frq ), % HERE
	atom_concat( Stem, '.mvisits', Fl ),
	atom_concat( Stem, '.x', FlPer ), 
	exp_loss_models_from_freqs( LFns, NofVs, FlPer, Frq ),
	min_exp_loss_naive_fs_ls( Fls, LFns ).

exp_loss_models_from_freqs( [], _N, _Stem, _Frq ).
exp_loss_models_from_freqs( [LossF|Fs], N, Stem, Frq ) :-
	frq_to_posterior_appx( Frq, N, Pappx ),
	atom_concat( pre_, LossF, PreLF ),
	( current_predicate( PreLF/2 ) ->
		pre_exp_loss_models( Pappx, PreLF, PreFrq )
		;
		PreFrq = Pappx
	),
	rel_exp_losses_for_fnc( PreFrq, LossF, RelLosses, MinLoss ),
	loss_id( LossF, FId ),
	atom_concat( Stem, FId, FnFile ),
	% here also record a matric of missmatch to Frq
	Terms = [min_loss(MinLoss)|RelLosses],
	write_terms( FnFile, write, Terms, [quoted(true)] ),
	exp_loss_models_from_freqs( Fs, N, Stem, Frq ).

rel_exp_losses_for_fnc( [PH-CMH|T], Loss, [HRL-MH|TRLs], MinLoss ) :-
	( CMH = (CH,MH) -> true ; CH = CMH, MH = CMH ),
	rel_exp_loss_of_mod( T, CH, Loss, 0, HRL ),
	rel_exp_losses_for_fnc_1( T, [PH-MH], Loss, HRL-MH, TRLs, MinLoss ).

rel_exp_losses_for_fnc_1( [], _Seen, _Loss, ML, [], ML ).
rel_exp_losses_for_fnc_1( [PH-CMH|T], Seen, Loss, AcRL-AcM, [HRL-MH|TRLs], ML ) :-
	( CMH = (CH,MH) -> true ; CH = CMH, MH = CMH ),
	rel_exp_loss_of_mod( Seen, CH, Loss, 0, MidHRL ),
	rel_exp_loss_of_mod( T, CH, Loss, MidHRL, HRL ),
	( HRL < AcRL -> NxRL is HRL , NxM = MH ; NxRL = AcRL, NxM = AcM ),
	rel_exp_losses_for_fnc_1( T, [PH-CH|Seen], Loss, NxRL-NxM, TRLs, ML ).

rel_exp_loss_of_mod( [], _RelMod, _Loss, ExpLoss, ExpLoss ).
rel_exp_loss_of_mod( [PH-CMH|T], RelMod, Loss, Acc, ExpLoss ) :-
	( CMH = (CH,MH) -> true ; CH = CMH, MH = CMH ),
	Call =.. [Loss,RelMod,CH,RLofH],
		% HERE, make sure of symmetricity and/or order
	call( Call ),
	NxAcc is Acc + (PH * RLofH),
	rel_exp_loss_of_mod( T, RelMod, Loss, NxAcc, ExpLoss ).

mvisits_file_to_frq_list( File, NofVs, Frq ) :-
	open( File, read, In ),
	read( In, First ),
	mvisits_stream_to_frq_list( First, In, 0, Frq, NofVs ),
	close( In ).

mvisits_stream_to_frq_list( end_of_file, _In, NofVs, [], NofVs ) :- !.
mvisits_stream_to_frq_list( Tms-c(Mod), In, Acc, [Tms-Mod|TFrqs], NofVs ) :-
	!,
	read( In, NxTerm ),
	NxAcc is Acc + Tms,
	mvisits_stream_to_frq_list( NxTerm, In, NxAcc, TFrqs, NofVs ).
mvisits_stream_to_frq_list( Term, In, Acc, Frq, NofVs ) :-
	write( user_error, 'mvisits_stream_to_frq_list/5: Skipping unrecognised term: ' ), write( user_error, Term ),
	nl( user_error ),
	write( user_error, '           in mvisits file. ' ), nl( user_error ),
	read( In, NxTerm ),
	mvisits_stream_to_frq_list( NxTerm, In, Acc, Frq, NofVs ).
	
pre_exp_loss_models( [], _Pred, [] ).
pre_exp_loss_models( [F-M|T], Pred, [F-(PreExM,M)|TPFs] ) :-
	Call =.. [Pred,M,PreExM],
	call( Call ),
	pre_exp_loss_models( T, Pred, TPFs ).

frq_to_posterior_appx( [], _N, [] ).
frq_to_posterior_appx( [F-M|T], N, [PH-M|R] ) :-
	PH is F / N,
	frq_to_posterior_appx( T, N, R ).
