/* 
     walk along a produced chain, displaying current/proposed model
     I should really synchronise this, to the code in elm('carts/auxil')
     */

% :- ensure_loaded( disp_bn ).     % /2.
:- ensure_loaded( edge_difs ).     % /5.
:- requires( disp_graph/2 ).
:- requires( kv_compose_fixed_v/3 ).
:- requires( kv_transpose/2 ).

:- ensure_loaded( library(readutil) ).     % read_line_to_codes/2
:- ensure_loaded( library(charsio) ).      % write_to_chars/2

disp_mcmc( File ) :-
     open( File, read, In ),
     read( In, Term ), 
     disp_mcmc( Term, 1, In ),
     close( In ).

disp_mcmc( eof, _, _In ) :- !.
disp_mcmc( end_of_file, _I, _In ) :- !.
disp_mcmc( Other, I, In ) :-
     !,
     ( Other = c(Current) ->
          NxtI is I + 1,
          read( In, b(Proposed) ), 
          edge_difs( Current, Proposed, Ce, Pe, Rev ),
          write( ce(Ce) ), nl,
          write( pe(Pe) ), nl,
          write( rev(Rev) ), nl,
          % disp_bn( Proposed ),
          % here
          term_to_dbl_quoted_atom( I, CurTitle ),
          atomic_concat( I, '*', Istar ),
          term_to_dbl_quoted_atom( Istar, PrpTitle ),
          % term_to_dbl_quoted_atom( I:c, CurTitle ),
          % term_to_dbl_quoted_atom( I:p, PrpTitle ),
          kv_compose_fixed_v( Ce, [color(blue)], OnlyInC ),
          kv_compose_fixed_v( Pe, [color(blue)], OnlyInP ),

          kv_transpose( Rev, Ver ),
          kv_compose_fixed_v( Ver, [color(brown)], RevInC ),
          kv_compose_fixed_v( Rev, [color(brown)], RevInP ),

          append( OnlyInC, RevInC, CClrs ),
          append( OnlyInP, RevInP, PClrs ),

          Copts = [edges_colours(CClrs), background(true), title(CurTitle) ],
          Popts = [edges_colours(PClrs), background(true), title(PrpTitle) ],
          disp_graph( Current, Copts ),
          disp_graph( Proposed, Popts ),
          % disp_bn( Proposed, Current ), 
          write( ' Press return to continue (0 first, to stop) : ' ),
          flush_output,
          read_line_to_codes( user_input, What ),
          % atom_codes( Atm, What ),
          % write( atom(Atm)-what(What) ), nl,
          ( What = [0'0] -> 
               flush_output, 
               flush_output(user_error),
               abort
               ; 
               write( user_error, What), nl(user_error)
          )
          ;
          NxtI = I
     ),
     read( In, Next ),
     disp_mcmc( Next, NxtI, In ).

term_to_dbl_quoted_atom( Term, Title ) :-
     write_to_chars( Term, NkdTitleCs ),
     append( [0'"|NkdTitleCs], [0'"], TitleCs ),
     atom_codes( Title, TitleCs ).
