:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( library(requires) ).
:- requires( to_list/2 ).
:- requires( directory_files_regular/2 ).
:- requires( file_stems_of_files_with_this_ext/3 ).
:- requires( werr/1 ).
:- requires( sys_info/1 ).
:- requires( nth_replace/5 ). 
:- requires( div_list/3 ). 
:- prolog_load_context( directory, Dir ), bb_put( juxtad, Dir ).

juxta_two_chains( OutAtomOr ) :-
	( OutAtomOr == [] -> 
		werr( [['Please provide a single filename for output.']] ),
		abort
		;
		to_list( OutAtomOr, [OutFile] )
	),
	directory_files_regular( '.', RegFiles ),
	file_stems_of_files_with_this_ext( RegFiles, '.gz', GzStems ),
	( GzStems = [File1,File2|TGzStems] ->
		true
		; 
		werr( [['Coulnd not locate two gzip files.']] ), abort
	),
	( TGzStems == [] -> 
		true 
		;
		werr( [['Ignoring files: '],['  '|TGzStems]] ), abort
	),
	gzip_file_to_parenthood_frq( File1, Frq1 ),
	gzip_file_to_parenthood_frq( File2, Frq2 ),
	write( done(File2) ), nl,
	open( OutFile, write, Out ),
	write_two_pa_frqs_for_plotting( Frq1, Frq2, Out ),
	close( Out ),
	bb_get( juxtad, Dir ),
	atom_concat( Dir, '/juxta_chains_plot ', JuxtaD ),
	atom_concat( JuxtaD, OutFile, Juxta ),
	sys_info( Juxta ).

write_two_pa_frqs_for_plotting( [], [], _Out ).
write_two_pa_frqs_for_plotting( [H1|T1], [H2|T2], Out ) :-
	write_two_lists_on_stream( H1, H2, Out ),
	write_two_pa_frqs_for_plotting( T1, T2, Out ).

write_two_lists_on_stream( [], [], _Out ).
write_two_lists_on_stream( [H1|T1], [H2|T2], Out ) :-
	write( Out, H1 ), write( Out, ' ' ), write( Out, H2 ), nl( Out ),
	write_two_lists_on_stream( T1, T2, Out ).

gzip_file_to_parenthood_frq( Stem, Frq ) :-
	atom_concat( Stem, '.gz', GzFile ),
	atom_concat( 'gunzip ', GzFile, Gunzip ),
	sys_info( Gunzip ),
	open( Stem, read, In ),
	read( In, First ),
	zero_parents( Zeros ),
	stream_count_parents( First, In, Zeros, 0, Counts, Tot ),
	normalise_parenthod_counts( Counts, Tot, Frq ),
	close( In ),
	atom_concat( 'gzip ', Stem, Gzip ),
	sys_info( Gzip ).

stream_count_parents( end_of_file, _In, Acc, N, Counts, Tot ) :-
	!,
	Counts = Acc,
	Tot is N.
stream_count_parents( Term, In, Acc, N, Counts, Tot ) :-
	( Term = c(Bn) -> 
		bn_adds_counts( Bn, Acc, NxAcc ),
		NxN is N + 1
		;
		Acc = NxAcc,
		NxN is N
	),
	read( In, NxTerm ),
	stream_count_parents( NxTerm, In, NxAcc, NxN, Counts, Tot ).

bn_adds_counts( [], Acc, Acc ).
bn_adds_counts( [Ch-Par|T], Acc, Out ) :-
	( Par == [] -> 
		Acc = MdAcc
		;
		nth_replace( Ch, Acc, New, Old, MdAcc ),
		parents_add_to_counts( Par, Old, New )
	),
	bn_adds_counts( T, MdAcc, Out ).

% parents_add_to_counts( +Par, +OldAccPars, New )
parents_add_to_counts( [], New, New ).
parents_add_to_counts( [H|T], Old, New ) :-
	nth_replace( H, Old, IncCntH, CntH, Nxt ),
	IncCntH is CntH + 1,
	parents_add_to_counts( T, Nxt, New ).


normalise_parenthod_counts( [], _Tot, [] ).
normalise_parenthod_counts( [List|T], Tot, [FrqL|FrqT] ) :-
	div_list( List, Tot, FrqL ),
	normalise_parenthod_counts( T, Tot, FrqT ).

zero_parents( [Z,Z,Z,Z,Z,Z,Z,Z] ) :-
	Z = [0,0,0,0,0,0,0,0].
