% given a stem (already ungzipped) chain file, it produces .llhoods, .leaves
% and .depts stats for that run.

:- ensure_loaded( library(system) ).		% file_exists/1.
:- ensure_loaded( '../../src/init_lib' ).			% library().
:- ensure_loaded( library(sys_info) ).		% /1

lld_stats_for_stem( [] ).
lld_stats_for_stem( [H|T] ) :-
	lld_stats_for_stem_1( H ),
	lld_stats_for_stem( T ).

lld_stats_for_stem_1( Stem ) :-
	atom_concat( Stem, '.llhoods', Llhoods ),
	( file_exists( Llhoods ) ->
		true
		;
		gunzip_if_stem( Stem ),
		atom_concat( Stem, '.leaves', Leaves ),
		atom_concat( Stem, '.depths', Depths ),
		open( Stem, read, In ),
		open( Llhoods, write, OutLl ),
		open( Leaves, write, OutLvs ),
		open( Depths, write, OutDpt ),
		read( In, c(Cart) ),
		read( In, x(c,Llhood,_Bpt,_XcMslf) ),
		read( In, BCart ),
		record_in_iterations( BCart, Cart, Llhood, In, OutLl, OutLvs, OutDpt ),
		close( In ), close( OutLl ), close( OutLvs ), close( OutDpt )
		% gzip_if_stem( Stem )
		% atom_concat( 'gzip ', Stem, Gzip ), sys_info( Gzip )
	).

gunzip_if_stem( Stem ) :-
	( file_exists(Stem) -> 
		true
		;
		atom_concat( Stem, '.gz', GzFile ),
		atom_concat( 'gunzip ', GzFile, Gunzip ),
		sys_info( Gunzip )
	).

record_in_iterations( end_of_file, Cart, Llhood, _In, Ll, Lvs, Dpt ) :-
	!,
	get_cart_s_nof_leaves_and_max_depth( Cart, 0, 0, NOfLeaves, MaxDepth ),
	write( Ll, Llhood ), nl( Ll ),
	write( Lvs, NOfLeaves), nl( Lvs ),
	write( Dpt, MaxDepth), nl( Dpt ).
record_in_iterations( b(_), Cart, Llhood, In, Ll, Lvs, Dpt ) :-
	!,
	get_cart_s_nof_leaves_and_max_depth( Cart, 0, 0, NOfLeaves, MaxDepth ),
	write( Ll, Llhood ), nl( Ll ),
	write( Lvs, NOfLeaves), nl( Lvs ),
	write( Dpt, MaxDepth), nl( Dpt ),
	read( In, XbOrC ),
	( XbOrC = x(b,_,_,_) -> 
		read( In, c(NxCart) )
		;
		XbOrC = c(NxCart)
	),
	read( In, x(c,NxLlhood,_,_) ),
	read( In, NxtTrm ),
	record_in_iterations( NxtTrm, NxCart, NxLlhood, In, Ll, Lvs, Dpt ).
record_in_iterations( _, Cart, Llhood, In, Ll, Lvs, Dpt ) :-
	read( In, NxtTrm ),
	record_in_iterations( NxtTrm, Cart, Llhood, In, Ll, Lvs, Dpt ).

get_cart_s_nof_leaves_and_max_depth( leaf(D,_I), Acc, _ManD, Cur, D ) :-
	!,
	% should have: ManD =:= D,
	Cur is Acc + 1.
get_cart_s_nof_leaves_and_max_depth( cart(_F,_C,L,R), Acc, ManD, Cur, D ) :-
	!,
	NxMD is ManD + 1,
	get_cart_s_nof_leaves_and_max_depth( L, Acc, NxMD, Mid, DL ),
	get_cart_s_nof_leaves_and_max_depth( R, Mid, NxMD, Cur, DR ),
	D is max( DL, DR ).
get_cart_s_nof_leaves_and_max_depth( l, Acc, ManD, Cur, D ) :-
	!,
	D = ManD,
	Cur is Acc + 1.
get_cart_s_nof_leaves_and_max_depth( cr(_F,_C,L,R), Acc, ManD, Cur, D ) :-
	NxMD is ManD + 1,
	get_cart_s_nof_leaves_and_max_depth( L, Acc, NxMD, Mid, DL ),
	get_cart_s_nof_leaves_and_max_depth( R, Mid, NxMD, Cur, DR ),
	D is max( DL, DR ).
	
