:- ensure_loaded( library(lists) ).					% memberchk/2.

:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( library(mcmcms_top_dir) ).

% :- elm( 'carts/slps/data_to_split_root' ).	% /3.
:- elm( 'auxil/dset_has_nofeatures' ).	% /2.
:- elm( 'carts/src/cart_classifies' ).
							% cart_classifies/7.
							% distribution_to_correct/4.

:- requires( opts_have_help/1 ).
:- requires( opts_select_infiles/3 ).
:- requires( write_list_elem_per_line/1 ).
:- requires( write_list_elem_per_line_w_period/1 ).
:- requires( select_n_uniformly/5 ).
:- requires( read_terms/2 ).
:- requires( kv_decompose_on_max_v/3 ).
:- requires( kv_multi_vs_by/3 ).
:- requires( kv_div_vs_by/3 ).
:- requires( kv_sum_vs/2 ).
:- requires( kv_decompose/3 ).
:- requires( write_terms/2 ).
% :- requires( elm/1 ).
:- requires( elm_location/2 ).

% 2005bb18. Cohesions became incremental. Code is likely to be more unstable.
% Example current usage can be found in elm('pbl/icml05/auxil/mk_predictions').
% Another addition is possibility of splitting probabilistic output to 
% two files: pbc_pn, splits the output to _pos and _neg part. The idea is that
% when plotting the predictions different mark points can be used for the two
% sets. A positive is a match between true class and positive class (the latter
% is either defined by the user with pstv/1, or the first Class on first predicted
% distro).
%
% upsh ~/islp/mcmcms/carts/auxil/test_classification train=pima_train test=pima_test method=ratio *mvisits 
% upsh ~/islp/mcmcms/carts/auxil/test_classification train=pima_train test=pima_test mvisits=tr_uc_rm_pima_idsd_train_a0_95b0_8_i50__s803.mvisits
% for fig.5 (ijcai05).
% upsh ~/islp/mcmcms/carts/auxil/test_classification train=pima_train test=pima_test mvisits=*mvisits method=bin

test_classification( Opts ) :-
	opts_have_help( Opts ),
	!,
	write( 'test_classification:     classify a number of test data points,' ),
	nl,
	write( '                         according to a training data-set and an' ),
	nl, 
	write( '                         fterms/mvisits file of an MCMC run.' ), nl, nl,
	write( '  Options.' ), nl,
	write( '    adj(A)               adjustment for leaf prediction (none,*bayes*)'), nl,
	write( '    cohese(C)            cohesion method: {best,*distro*}' ), nl,
	write( '    file(M)              the files to process' ), nl,
	write( '    output(O)            output method: {bin,*pbc*,pbc_pn}' ), nl,
	write( '    pstv(Pstv)           positive class (defaults to first)' ), nl,
	write( '    term(B)              term period at true' ), nl,
	write( '    test(S)              test data-set file' ), nl,
	write( '    test_data(D)         the test data: (no default)' ), nl,
	write( '    train(N)             the training data-set.' ), 
     write( '    train_ids(I)         the train ids (no default)' ), nl,
	write( '    sfx(Sfx)             suffix for extension of each file' ), nl,
     nl,
     write( '    Notes: If test_data/1 is present test/1 is not used.' ),
     write( '           If train_ids/1 is present train/1 is not used' ),
     nl, nl, nl.

test_classification( Opts ) :-
     ( memberchk(train_ids(Ids),Opts) ->
          TrainOpts = []
          ;
	     memberchk( train(Train), Opts ), 
	     once( train_loaded_and_calls( Train, Ids, TrainOpts ) )
     ),
	% findall( Mvs, member(mvisits(Mvs),Opts), MVsFs ), 
	opts_select_infiles( Opts, MVsFs, RmOpts ),
	write( files(MVsFs) ), nl,
	append( RmOpts, TrainOpts, TstClassOpts ),
     ( memberchk(test_data(TstData),Opts) -> true
	     ;
	     memberchk( test(Test), Opts ), 
          test_loaded( Test, TstData )
     ),
	test_classification_1( MVsFs, Ids, TstData, TstClassOpts ).

test_classification_1( [], _Ids, _TstData, _Opts ).
test_classification_1( [Stem|T], Ids, TstData, Opts ) :-
	write( doing(Stem) ), nl,
	( memberchk( term(TBoo), Opts ) -> true; TBoo = true ),
     atom_concat( Stem, '.fterms', Mvs ),
     atom_concat( Stem, '.gterms', Gtr ),
     % atom_concat( Stem, '.mvisits', Mvs ),
	open( Mvs, read, In ), 
	open( Gtr, write, Out ), 
	read( In, InTerm ),
	( memberchk(cohese(CohM),Opts) -> true; CohM = distro ), % alt. Meth= best
	initial_cohesion_accumulator( TstData, AccDs ),
	length( TstData, NofTD ),
	TstData = [HDatum|_],
	functor( HDatum, _, DAr ),
	% current_prolog_flag( gc, Gc ), 
	% write( user_error , current_prolog_flag(gc,Gc) ), nl( user_error ),
	(memberchk(adj(Adj),Opts) -> true ; Adj = bayes ),
	stream_mvis_terms_to_classified( InTerm, 1, Ids, TstData, DAr, NofTD, Adj, CohM, 0, 0/0/BinAcrTrm/ExpAcrTrm, In, AccDs, ITms, CohDs, Out ),
	close( In ),
	close( Out ),
	write( done_with_mvis ), nl,
	normalise_cohesed( CohDs, ITms, NrmDs ),
	write( done_with_normalising ), nl,
	( memberchk(output(OutM),Opts) -> true; OutM = pbc ), % alt. = bin, pbc_pn
	( memberchk( sfx(AllSfxPrv), Opts ) -> 
		( atom_concat( '_', _, AllSfxPrv ) ->
			AllSfxPrv = AllSfx
			;
			atom_concat( '_', AllSfxPrv, AllSfx )
		)
		; AllSfx = '' ),
	% cohese_class_pairs( Distros, TrueClasses, CohM, OutM, ITms, Cohesed ), 
	% kv_decompose( CohesedPrs, Cohesed, _ ),
	% atom_concat( Stem, '.fterms', Mvs ),
	% atom_concat( Stem, '.mvisits', Mvs ),
	( (CohM == distro, (OutM == pbc; OutM == pbc_pn)) ->
		Sfx = '.predicted'
		;
		atom_concat( '.predicted_', CohM, Sfx1 ),
		atom_concat( '_', OutM, Sfx2 ),
		atom_concat( Sfx1, Sfx2, Sfx )
	),
	atom_concat( Stem, Sfx, TstClsPrv ),
	atom_concat( TstClsPrv, AllSfx, TstCls ),
	( OutM == pbc_pn -> 
		( memberchk(pstv(Pstv),Opts) -> true; 
			NrmDs = [[Pstv-_|_]|_]
		),
		to_file_period_opt( TstCls, Adj, TBoo, NrmDs ),
		data_to_true_classes( TstData, TrCls ), % here
		separate_to_pos_and_negative( NrmDs, TrCls, Pstv, PstvDs, NgtvDs ),
		atom_concat( TstCls, '_pos', TstPstv ),
		atom_concat( TstCls, '_neg', TstNgtv ),
		to_file_period_opt( TstPstv, Adj, TBoo, PstvDs ),
		to_file_period_opt( TstNgtv, Adj, TBoo, NgtvDs )
		;
		to_file_period_opt( TstCls, Adj, TBoo, NrmDs )
	),
	atom_concat( '.predicted', MethsId, Sfx ),
	atom_concat( '.accuracy_bin', MethsId, AcrBinSfx ),
	atom_concat( '.accuracy_exp', MethsId, AcrExpSfx ),
	atom_concat( Stem, AcrBinSfx, AcrBinPrv ),
	atom_concat( AcrBinPrv, AllSfx, AcrBin ),
	atom_concat( Stem, AcrExpSfx, AcrExpPrv ),
	atom_concat( AcrExpPrv, AllSfx, AcrExp ),
	% count_accuracy( Cohesed, TstData, DAr, bin, 0, 0, BinAcrTrm ),
	% count_accuracy( Cohesed, TstData, DAr, exp, 0, 0, ExpAcrTrm ),
	% count_accuracy( TrueTuples, bin, 0, 0, BinAcrTrm ),
	% count_accuracy( TrueTuples, exp, 0, 0, ExpAcrTrm ),
	write_terms( AcrBin, [BinAcrTrm] ),
	write_terms( AcrExp, [ExpAcrTrm] ),
	test_classification_1( T, Ids, TstData, Opts ).

separate_to_pos_and_negative( [], [], _Pstv, [], [] ).
separate_to_pos_and_negative( [H|T], [TrClass|TrClassesT], Pstv, PstvDs, NgtvDs ) :-
	% kv_decompose_on_max_v( H, [BstKat|_], _MaxV ),
	( TrClass == Pstv ->
		PstvDs = [H|TPstvDs],
		NgtvDs = TNgtvDs
		;
		PstvDs = TPstvDs,
		NgtvDs = [H|TNgtvDs]
	),
	separate_to_pos_and_negative( T, TrClassesT, Pstv, TPstvDs, TNgtvDs ).

% count_accuracy( Cohesed, Data, DAr, Whc, AccSum, AccTot, AcrTrm ) :-
count_accuracy( [], _Data, _DAr, _Whc, AccSum, AccTot, AcrTrm ) :-
	AcrTrm = AccSum/AccTot.
count_accuracy( [HDst|TDsts], [HDtm|TDt], DAr, Whc, AccS, AccT, Acr ) :-
	arg( DAr, HDtm, TrueClass ),
	memberchk( TrueClass-TruePrb, HDst ),
	distribution_to_correct( HDst, TruePrb, TrueClass, Corr ),
	( Whc == bin ->
		( Corr == true ->
			NxSum is AccS + 1
			;
			NxSum is AccS
		)
		;
		NxSum is AccS + TruePrb
	),
	NxTot is AccT + 1,
	count_accuracy( TDsts, TDt, DAr, Whc, NxSum, NxTot, Acr ).

% stream_mvis_terms_to_classified( end_of_file, _TrnD, _TstD, Tot, _In, [], Tot ) :- !.
stream_mvis_terms_to_classified( end_of_file, _I, _Ids, _TstD, _DAr, _NofTD, _Adj, _CohM, Tot, Acr, _In, CohDs, Tot, CohDs, _Out ) :-
	!,
	Acr = BinAcrSum/ExpAcrSum/BinAcrTrm/ExpAcrTrm,
	BinAcrTrm = BinAcrSum/Tot, % not 'is' instead of =
	ExpAcrTrm = ExpAcrSum/Tot.
% stream_mvis_terms_to_classified( Tms-c(Cart), I, Ids, TstData, DAr, NofTD, Adj, CohM, Acc, AccAcr, In, AccCohDistros, Tot, CohDs ) :-
stream_mvis_terms_to_classified( st(Tms,Llh,Mis,Cart), I, Ids, TstData, DAr, NofTD, Adj, CohM, Acc, AccAcr, In, AccCohDistros, Tot, CohDs, Out ) :-
%% stream_mvis_terms_to_classified( Tms-c(Cart), TrainData, TstData, DAr, NofTD, CohM, Acc, AccAcr, In, AccCohDistros, Tot, CohDs ) :-
% stream_mvis_terms_to_classified( Tms-c(Cart), TrainData, TstData, Acc, In, [Distros-Tms|TPairs], Tot ) :-
	% data_to_split_root( TrainData, _Splits, Ids ),
	cart_classifies( Cart, Ids, TstData, DAr, Adj, Distros, CorrPrs ),
	cohese_incremental( CohM, Distros, AccCohDistros, Tms, NxAccCohDistros ),
	AccAcr = AccBinAcr/AccExpAcr/BinAcr/ExpAcr,
	sum_new_accuracies( CorrPrs, 0, 0, CartBinSum, CartExpSum ),
     keysort( CorrPrs, SrtPrs ),
     per_cart_per_class_stats( SrtPrs, PcPc, Err ),
	NxAccBinAcr is AccBinAcr + ((CartBinSum/NofTD) * Tms),
	NxAccExpAcr is AccExpAcr + ((CartExpSum/NofTD) * Tms),
	NxAccAcr = NxAccBinAcr/NxAccExpAcr/BinAcr/ExpAcr,
	% write( '.' ), nl,
	NxAcc is Acc + Tms,
	read( In, InTerm ),
	!,
	( I mod 5000 =:= 0 -> 
		( pl(sicstus(_)) -> 
			write( user_error, I ), write( user_error, ' : ' ),
			write( user_error, forcing_gc ), garbage_collect,
			write( user_error, '...' ), garbage_collect_atoms,
			write( user_error, ok ), nl( user_error )
			;
			write( user_error, I ), write( user_error, ' ' ),
               nl( user_error )
		)
		;
		true
	),
	NxI is I + 1,
     AddMis = [Err|PcPc],
     ( Mis = [_|_] -> append(Mis,AddMis,NewMis) ; NewMis=[Mis|AddMis] ),
     writeq( Out, st(Tms,Llh,NewMis,Cart) ), write( Out, '.' ), nl( Out ),

	stream_mvis_terms_to_classified( InTerm, NxI, Ids, TstData, DAr, NofTD, Adj, CohM, NxAcc, NxAccAcr, In, NxAccCohDistros, Tot, CohDs, Out ).
	%% stream_mvis_terms_to_classified( InTerm, TrainData, TstData, DAr, NofTD, CohM, NxAcc, NxAccAcr, In, NxAccCohDistros, Tot, CohDs ).

sum_new_accuracies( [], Bin, Exp, Bin, Exp ).
sum_new_accuracies( [_-(Prb:Bool)|T], AccBin, AccExp, Bin, Exp ) :-
	( Bool == true -> 
		NxBin is AccBin + 1
		;
		NxBin = AccBin
	),
	NxExp is AccExp + Prb,
	sum_new_accuracies( T, NxBin, NxExp, Bin, Exp ).

per_cart_per_class_stats( [K-(Prb:Bool)|T], PcPc, Err ) :-
     bool_to_add_val( Bool, Val ),
     AcSm is (1 - Prb) ** 2,
     per_cart_per_class_stats_1( T, K, Val, Prb, AcSm, 1, 1, PcPc, Err ).

per_cart_per_class_stats_1( [], K, AcTms, AcPrb, Sum, Ct, Lt, PcPc, Err ) :-
     PcPc = [K:(AcTms#AcPrb#Ct)],
     Err  is sqrt( Sum/Lt ).
per_cart_per_class_stats_1( [K-Str|T], K, AcTms, AcPrb, AcSm, Ct, Lt, PcPc, Err ) :-
     !,
     Str = Prb:Bool,
     bool_to_add_val( Bool, Val ),
     NxTms is AcTms + Val,
     NxPrb is AcPrb + Prb,
     NxSm is AcSm + ((1-Prb)**2),
     NxCt is Ct + 1,
     NxLt is Lt + 1,
     per_cart_per_class_stats_1( T, K, NxTms, NxPrb, NxSm, NxCt, NxLt, PcPc, Err ).
per_cart_per_class_stats_1( [K-Str|T], OldK, AcTms, AcPrb, AcSm, Ct, Lt, PcPc, Err ) :-
     PcPc = [OldK:(AcTms#AcPrb#Ct)|TPcPc],
     Str = Prb:Bool,
     bool_to_add_val( Bool, Val ),
     NxSm is AcSm + ((1-Prb)**2),
     NxLt is Lt + 1,
     per_cart_per_class_stats_1( T, K, Val, Prb, NxSm, 1, NxLt, TPcPc, Err ).

bool_to_add_val( true, 1 ).
bool_to_add_val( false, 0 ).

cohese_incremental( CohM, IncDistros, Acc, Tms, NxAcc ) :-
	(CohM == best -> 
		cohese_inc_best( IncDistros, Acc, Tms, NxAcc )
		;
		CohM == distro,
		cohese_inc_distro( IncDistros, Acc, Tms, NxAcc )
	).

%%% HERE
cohese_inc_best( [], [], _Tms, [] ).
cohese_inc_best( [Hinc|Tincs], [Hacc|Taccs], Tms, [Hnx|Tnxs] ) :-
	% best_in_distro( Hinc, BstKat ), =>
	kv_decompose_on_max_v( Hinc, [BstKat|_], _MaxV ),
	kvs_replace_v( Hacc, BstKat, Old, New, Hnx ),
	New is Old + Tms,
	cohese_inc_best( Tincs, Taccs, Tms, Tnxs ).

cohese_inc_distro( [], [], _Tms, [] ).
cohese_inc_distro( [Hinc|Tincs], [Hacc|Taccs], Tms, [Hnx|Tnxs] ) :-
	cohese_distros( Hinc, Hacc, Tms, Hnx ),
	cohese_inc_distro( Tincs, Taccs, Tms, Tnxs ).

cohese_distros( [], [], _Tms, [] ).
cohese_distros( [Hkat-HPrb|T], [Hkat1-HAccPrb|Taccs], Tms, [Hkat-HnxPrb|Tnxs] ) :-
	HnxPrb is HAccPrb + (HPrb * Tms),
	( Hkat == Hkat1 -> true ; 
	    write( user_error, missmatch_kategories(Hkat,Hkat1) ), 
	    nl( user_error ), abort
	),
	cohese_distros( T, Taccs, Tms, Tnxs ).

cohese_class_pairs( [[]-_|_], _Empty, _CohM, _OutM, _Its, [] ) :- !.
cohese_class_pairs( Pairs, [HCls|TTC], CohM, OutM, Its, [ThisRec|Rest] ) :-
	kv_skim_ks( Pairs, Froth, Rem ),
	(CohM == best -> 
		distros_to_best_classes( Froth,  Best ),
		kv_on_k_add_vs( Best, SummedUnoInc ),
		Froth = [HFro-_|_],
		kv_adds_def_v_where_missing( HFro, SummedUnoInc, 0, SummedUno ),
		keysort( SummedUno, Summed ),
		( OutM== bin -> 
			kv_decompose_on_max_v( Summed, MaxClasses, _MaxV ),
			length( MaxClasses, Length ),
			select_n_uniformly( 1, MaxClasses, Length, [ThisRec], _ )
			% select_n_uniformly( 1, MaxClasses, Length, [ThisClass], _ ),
			% method_to_record_point( Meth, ThisClass, MaxV, Its, ThisRec ),
			;
			kv_sum_vs( Summed, VsSum ),
			kv_div_vs_by( Summed, VsSum, ThisRec )
		)
		;
		Froth = [HFro-Tms|TFro],
		kv_multi_vs_by( HFro, Tms, AdjHFro ),
		cohese_datum_distros( TFro, AdjHFro, CohDistro ),
		( OutM == bin ->
			kv_decompose_on_max_v( CohDistro, MaxClasses, _MaxV ),
			length( MaxClasses, Length ),
			select_n_uniformly( 1, MaxClasses, Length, [ThisRec], _ )
			;
			kv_sum_vs( CohDistro, VsSum ),
			kv_div_vs_by( CohDistro, VsSum, ThisRecPbc ),
			( OutM == pbc ->
				ThisRec = ThisRecPbc
				;
				OutM == pbc_pn,
				ThisRec = ThisRecPbc-HCls
			)
		)
	),
	cohese_class_pairs( Rem, TTC, CohM, OutM, Its, Rest ).

% cohese_datum_distros( +Pairs, +AccDistro, -Distro ) :-
% Distro is unormalised.
%
cohese_datum_distros( [], Distro, Distro ).
cohese_datum_distros( [H-Tms|T], AccDistro, Distro ) :-
	kv_multi_vs_by( H, Tms, Contr ),
	add_vs( AccDistro, Contr, NxDistro ),
	cohese_datum_distros( T, NxDistro, Distro ).

method_to_record_point( class, Class, _MaxTms, _Its, Class ).
method_to_record_point( ratio, _Class, MaxTms, Its, MaxTms/Its ).
method_to_record_point( both, Class, MaxTms, Its, Class:MaxTms/Its ).

kv_adds_def_v_where_missing( [], KVs, _Def, KVs ).
kv_adds_def_v_where_missing( [K-_V|T], AccKVs, Def, KVs ) :-
	( memberchk(K-_V2,AccKVs) ->
		AccKVs = NxAccKVs
		;
		NxAccKVs = [K-Def|AccKVs]
	),
	kv_adds_def_v_where_missing( T, NxAccKVs, Def, KVs ).


kv_on_k_add_vs( [], [] ).
kv_on_k_add_vs( [Hk-Hv|T], Summed ) :-
	( select(Hk-OthV,T,Rem) ->
		NxV is OthV + Hv,
		NexT = [Hk-NxV|Rem],
		TSummed = Summed
		;
		NexT = T,
		Summed = [Hk-Hv|TSummed]
	),
	kv_on_k_add_vs( NexT, TSummed ).

kv_skim_ks( [], [], [] ).
kv_skim_ks( [[H|TKs]-V|TPrs], [H-V|Rst], [TKs-V|TRem] ) :-
	kv_skim_ks( TPrs, Rst, TRem ).

add_vs( [], [], [] ).
add_vs( [K-V1|T1], [K-V2|T2], [K-Sum|TSums] ) :-
	Sum is V1 + V2,
	add_vs( T1, T2, TSums ).

distros_to_best_classes( [],  [] ).
distros_to_best_classes( [H-Tms|T],  [HBst-Tms|TBst] ) :-
	kv_decompose_on_max_v( H, MaxKs, _MaxV ),
	length( MaxKs, Length ),
	select_n_uniformly( 1, MaxKs, Length, [HBst], _ ),
	distros_to_best_classes( T, TBst ).

train_loaded_and_calls( pima_train, Ids, [] ) :-
	elm( 'carts/pima/data/pima_idsd_train.pl' ), 
	% NofF is 8,
	pima_idsd_train( _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ).
train_loaded_and_calls( icml_pima, Ids, [] ) :-
	elm( 'pbl/icml05/data/idsd' ), 
	NofF is 8,
	idsd( NofF, pima_train_080, _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ),
train_loaded_and_calls( icml_bcw, Ids, [] ) :-
	elm( 'pbl/icml05/data/idsd' ), 
	NofF is 9,
	idsd( NofF, bcw_train_080, _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ).
train_loaded_and_calls( icml_kyph, Ids, [] ) :-
	elm( 'pbl/icml05/data/idsd' ), 
	NofF is 3,
	idsd( NofF, kyph_train_080, _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ).
train_loaded_and_calls( icml_letrec, Ids, [pstv('A')] ) :-
	elm( 'pbl/icml05/data/idsd' ), 
	NofF is 16,
	idsd( NofF, letter_recognition_train_080, _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ).
% train_loaded_and_calls( icml_wform, TrainData, [pstv(0)] ) :-
train_loaded_and_calls( icml_wform, Ids, [] ) :-
	elm( 'pbl/icml05/data/idsd' ), 
	NofF is 40,
	idsd( NofF, waveform_noise_data_train_080, _Splits, Ids ).
	% nof_features_to_train_data( NofF, TrainData ). % delete those after testing
train_loaded_and_calls( pima_ijcai05_synth_500, Ids, [] ) :-
	elm( 'pbl/ijcai05/data/idsd' ),
	% ensure_loaded( '$HOME/islp/mcmcms/pbl/ijcai05/data/idsd' ), 
	idsd( 8, sg500, _Splits, Ids ).
	% findall( [A,B,C,D,E,F,G,H]-I, data(A,B,C,D,E,F,G,H,I), TrainData ).

train_loaded_and_calls( sn1_trn_080, Ids, [] ) :-
	elm( 'carts/sn/data/idsd' ),
	idsd( 18, sn_clean1_trn_080, _Splits, Ids ).

train_loaded_and_calls( SN, Ids, [] ) :-
	atom_concat( sn_, _, SN ),
	write( train(SN) ), nl,
	elm( 'carts/sn/data/idsd' ),
	/*
	SNPrs = [sn_clean2_trn_080-19,sn_clean3_trn_080-33,
	         sn_clean4_trn_080-8, sn_clean5_trn_080-15],
	( memberchk(SN-NofF,SNPrs) -> 
		idsd( NofF, SN, _Splits, Ids )
		;
		throw( unknown_sn(SN) )
	).
	*/
	elm_location( 'carts/sn/data/', DataDir ),
	atom_concat( DataDir, SN, SNFile ),
	dset_has_nofeatures( SNFile, NofF ),
	idsd( NofF, SN, _, Ids ).

/*
nof_features_to_train_data( NofF, TrainData ) :-
	NofArgs is NofF + 1,
	functor( Call, data, NofArgs ),
	data_call_and_feature_list( NofF, NofArgs, Call, FList, Kat ),
	findall( FList-Kat, Call, TrainData ).
*/

test_loaded( pima_test, TestData ) :-
	!,
	bb_get( mcmcms_top_dir, Dir ),
	atom_concat( Dir, 'carts/pima/data/pima_test.pl', TstDFile ),
	read_terms( TstDFile, TestData ).
test_loaded( sn1_tst_020, TestData ) :-
	!,
	bb_get( mcmcms_top_dir, Dir ),
	atom_concat( Dir, 'carts/sn/data/sn_clean1_tst_020.pl', TstDFile ),
	read_terms( TstDFile, TestData ).
test_loaded( SN, TestData ) :-
	write( test(SN) ), nl,
	atom_concat( sn_, _, SN ),
	!,
	bb_get( mcmcms_top_dir, Dir ),
	atom_concat( Dir, 'carts/sn/data/', DtDir ),
	atom_concat( SN, '.pl', SNpl ),
	atom_concat( DtDir, SNpl, TstDFile ),
	read_terms( TstDFile, [FstDatum|TTestData] ),
	( functor(FstDatum,data,_) -> 
		TestData = [FstDatum|TTestData] 
		;
		TestData = TTestData
	).
test_loaded( File, TestData ) :-
	read_terms( File, TestData ).

to_file_period_opt( File, Adj, TBoo, Cohesed ) :-
	atom_concat( '_', Adj, UndAdj ),
	atom_concat( File, UndAdj, FileAdj ),
	tell( FileAdj ),
	( TBoo == false ->
		write_list_elem_per_line( Cohesed )
		;
		write_list_elem_per_line_w_period( Cohesed )
	),
	told.

initial_cohesion_accumulator( TData, Acc ) :-
	bb_get( all_categories, Kats ),
	kv_compose_fixed_v( Kats, 0, EmptyD ),
	initial_cohesion_accumulator_1( TData, EmptyD, Acc ).

initial_cohesion_accumulator_1( [], _EmptyD, [] ).
initial_cohesion_accumulator_1( [_|T], EmptyD, [EmptyD|TEDs] ) :-
	initial_cohesion_accumulator_1( T, EmptyD, TEDs ).

normalise_cohesed( [], _Itms, [] ).
normalise_cohesed( [H|T], Itms, [SclH|Tscl] ) :-
	kv_div_vs_by( H, Itms, SclH ),
	normalise_cohesed( T, Itms, Tscl ).


data_to_true_classes( [H|T], [Fst|Rst] ) :-
	functor( H, data, NofArgs ),
	arg( NofArgs, H, Fst ),
	data_to_true_classes_1( T, NofArgs, Rst ).

data_to_true_classes_1( [], _NofArgs, [] ).
data_to_true_classes_1( [H|T], NofArgs, [Fst|Rst] ) :-
	arg( NofArgs, H, Fst ),
	data_to_true_classes_1( T, NofArgs, Rst ).
