:- op( 600, xfy, ~ ).
:- ensure_loaded( '../../../../src/init_lib' ).		%.
:- ensure_loaded( '../../../../bns/auxil/disp_bn' ).
:- ensure_loaded( library(charsio) ).		% write_to_chars/2
:- ensure_loaded( library(delete_file_if) ).
:- ensure_loaded( library(lists) ).		% memberchk/2.

display_model( ModCart, Opts ) :-
	% write( this_is_where_we_display_bt(BT) ), nl.
	memberchk( title(Title), Opts ),
	memberchk( lhood(Llhood), Opts ),
	cart_to_structure( ModCart, 1, _, _, Str ),
	mktemp( tmp_disp_XXXXXX, Tmp ),
	cart_str_to_disp_dag( Str, Llhood, Title, Tmp ),
	atom_concat( Tmp, ' &', TmpSfx ),
	atom_concat( 'doted ', TmpSfx, Doted ),
	shell( Doted ),
	( bb_get(not_first_display_model,true) ->
		sleep(1) ; (sleep(4),bb_put(not_first_display_model,true)) ),
	delete_file_if( Tmp ).

cart_to_structure( cart(F,S,L,R), N, N:F:S, N3, [L1-N,R1-N,N:F:S|T] ) :-
	N1 is N + 1,
	cart_to_structure( L, N1, L1, N2, LStr ),
	cart_to_structure( R, N2, R1, N3, RStr ),
	append( LStr, RStr, T ).
cart_to_structure( leaf(_SD,SIds), N, N:Pop, NxN, [N:Pop] ) :-
	NxN is N + 1,
	( current_predicate(has_category/2) ->
		bb_get( all_categories, Kats ),
		mold_list( Kats, 0, Zeros ),
		leaf_ids_classified( SIds, Kats, Zeros, Classified ),
		classified_to_slashed( Classified, Pop )
		;
		length( SIds, Pop )
	).

mold_list( [], _E, [] ).
mold_list( [_|T], E, [E|R] ) :-
	mold_list( T, E, R ).

leaf_ids_classified( [], _Kats, Classified, Classified ).
leaf_ids_classified( [H|T], Kats, Accs, Classified ) :-
	has_category( H, HClass ),
	nth( Nth, Kats, HClass ), 
	nth_with_new( Nth, Accs, Old, New, NxAccs ), 
	New is Old + 1,
	leaf_ids_classified( T, Kats, NxAccs, Classified ).

nth_with_new( 1, [H|T], H, New, [New|T] ).
nth_with_new( Nth, [H|T], Old, New, [H|R] ) :-
	NxNth is Nth - 1,
	nth_with_new( NxNth, T, Old, New, R ).

classified_to_slashed( [F,S|T], F/R ) :-
	!,
	classified_to_slashed( [S|T], R ).
classified_to_slashed( [E], E ).

cart_str_to_disp_dag( Str, Lk, Frq, Dag ) :-
	pick_parents( Str, Parents, Leaves, Pairs, Edges ),
	keysort( Pairs, Sorted ),
	edges_parents_to_dag( Parents, Leaves, Sorted, Edges, Lk, Frq, Dag ).

% edges_parents_to_dag( [Left,Right|T], Parents, Dag ) :-
edges_parents_to_dag( Parents, Leaves, Pairs, Edges, Lk, Frq, Dag ) :-
	open( Dag, write, S ),
	write( S, 'digraph G {' ), nl( S ),
	write( S, '  node [style=filled]' ), nl( S ),
	write( S, '  node [shape=egg]' ), nl( S ),
	write( S, '  edge [color=red]' ), nl( S ),
	% write( S, '  1 [color=red]' ), nl( S ),
	leave_parent_nodes( Pairs, S ),
	% leave_nodes( Leaves, S ),
	% parent_nodes( Parents, S ),
	edges_w_labels( Edges, Parents, Leaves, S ),
	write( S, '"' ), write( S, Lk ), write( S, ' \\n ' ),
	write( S, Frq ), write( S, '"' ),
	write( S, '[color=lemonchiffon2, shape=invhouse]' ), nl( S ),
	write( S, '}' ),
	close( S ).

leave_parent_nodes( [], _S ).
leave_parent_nodes( [Id-Node|T], S ) :-
	( Id == Node ->
		dbl_quote_protect_pl_term( Node, QNode ),
		write( S, '  ' ), write( S, QNode ), 
		write( S, ' [color=green]' ), nl( S )
		;
		( Id =:= 1 ->
			write( S, '  ' ), write( S, Node ),
			write( S, ' [color=red]' ), nl( S )
			;
			write( S, '  ' ), write( S, Node ), 
			write( S, ' [color=orange]' ), nl( S )
		)
	),
	leave_parent_nodes( T, S ).
	

parent_nodes( [], _S ).
parent_nodes( [N-LLbl|T], S ) :-
	( N =:= 1 ->
		write( S, '  ' ), write( S, LLbl ),
		write( S, ' [color=red]' ), nl( S )
		;
		write( S, '  ' ), write( S, LLbl ), 
		write( S, ' [color=orange]' ), nl( S )
	),
	parent_nodes( T, S ).

leave_nodes( [], _S ).
leave_nodes( [H|T], S ) :-
	write( S, '  "' ), write( S, H ), 
	write( S, '" [color=green]' ), nl( S ),
	leave_nodes( T, S ).

edges_w_labels( [], _Parents, _Leaves, _S ).
edges_w_labels( [LCh-Par,RCh-Par|T], Parents, Leaves, S ) :-
	memberchk( Par-ParLbl, Parents ),
	write( S, '  ' ), write( S, ParLbl ),
	% ( nth( NL, Leaves, LCh ) -> true ; NL = LCh ),
	( cart_parent( LCh, _-LCs ) -> true ;
		dbl_quote_protect_pl_term( LCh, LCs )
	),
	write( S, ' -> ' ), write( S, LCs ), nl( S ),
	% write( S, ' [label=' ), 
	% select( Par-LLbl, Parents, NxParents ),
	% write( S, LLbl ), write( S, ']' ), nl( S ),
	write( S, '  ' ), write( S, ParLbl ),
	% ( nth( NR, Leaves, RCh ) -> true ; NR = RCh ),
	% ( cart_parent( RCh, _-PrpRCh ) -> true ; PrpRCh = RCh ),
	( cart_parent( RCh, _-RCs ) -> true ;
		dbl_quote_protect_pl_term( RCh, RCs )
	),
	write( S, ' -> ' ), write( S, RCs ), nl( S ),
	edges_w_labels( T, Parents, Leaves, S ).

pick_parents( [], [], [], [], [] ).
pick_parents( [H|T], Par, Leaves, Pairs, Edges ) :-
	( cart_edge(H,HEdge) ->
		Par = TPar,
		Leaves = TLeaves,
		Pairs = TPairs,
		Edges = [HEdge|TEdges]
		;
		( cart_parent(H,ParH) ->
			Par = [ParH|TPar],
			Leaves = TLeaves,
			Pairs = [ParH|TPairs],
			Edges = TEdges
			;
			Par = TPar,
			Leaves = [H|TLeaves],
			Pairs = [H-H|TPairs],
			Edges = TEdges
		)
	),
	pick_parents( T, TPar, TLeaves,TPairs, TEdges ).

cart_edge( Child-Parent, Child-Parent ).
	% ( cart_parent( Child, _N-EdgeCh ) -> true
		% ; EdgeCh = Child ),
	% ( cart_parent( Parent, _N-EdgePar ) -> true
		% ; EdgePar = Parent ).

cart_parent( N:F:Ch, N-QH ) :-
	% number_codes( N, Ns ), 
	% number_codes( F, Fs ),
	dbl_quote_protect_pl_term( N:F~Ch, QH ).
	% append( [0'_|Fs], [0'_|ChCs], SfxCs ),
	% append( Ns, SfxCs, ParHCs ),

dbl_quote_protect_pl_term( Term, Q ) :-
	write_to_chars( Term, TCs ),
	append( [0'"|TCs], [0'"], QCs ),
	atom_codes( Q, QCs ).
