:- ensure_loaded( edit_mod ). 			% variable_has_position_in_cart/3,edcart_to_concrete_cart/2.
:- ensure_loaded( library(lists) ).		% select/3.
:- ensure_loaded( library(ordsets) ).		% ord_union/3.

% EdCart = Editable Cart = ec(Cart,LeavesList,NodePositions).
% LeavesList = [leaf(Depth,Ids,Possibles)-LeafV|T]
% NodePositions = [[],[{l,r}*]*] l-left, r-right. a list for each internal node should be present
% Cart = cart(Feature, ChoiceOnFeature, Possibles, LeftChild, RightChild ).
%		if Left/RightChild is a leaf, then it will be a variable, same as corresponding LeafV in LeavesList
%         this representation adds Possibles to node levels when compared to growing slp.
% Leaf = leaf(Depth,Ids,Possibles).
%
9/10 :: move( CartIn, CartOut ) :- edit( CartIn, CartMid ), move( CartMid, CartOut ).
1/10 :: move( CartIn, CartOut)  :- edcart_to_concrete_cart( CartIn, CartOut ).

1/3 :: edit( CartIn, CartOut ) :-  				% grow
			CartIn = ec(Cart,Lvs,NdPsts),
			pmember(Lvs,Leaf,RmLvs),
			Leaf = leaf(D,Ids,Poss)-CartVar,
			variable_has_position_in_cart( Cart, CartVar, NwNodePos ),
			split_leaf( D, Ids, Poss, CartVar, RmLvs, NwLvs ),
			CartOut = ec(Cart,NwLvs,[NwNodePos|NdPsts] ).
1/3 :: edit( CartIn, CartOut ) :- 					% prune
			CartIn = ec(Cart,Lvs,NdPsts),
			prunable_places( Cart, [], PrnblPsts, [] ),
			pmember( PrnblPstsVs, PrunePstVs, _RemPrnbls ),
			navigate( Cart, PrunePst, PruneNode, PrunedCart, ConnctVar ),
			PruneNode = cart(_PrnF,_PrnCs,PrnPoss,LVar,RVar),
			leaves_list_var_select( Lvs, LVar, LLv, LrmLvs ),
			leaves_list_var_select( LrmLvs, RVar, RLv, RrmLvs ),
			select( PrnPoss, NdPsts, NewNdPsts ),
			% LLv = leaf(VanishDepth,_,_),
			LLv = leaf(LD,LIds,_LPoss),
			RLv = leaf(_RD,RIds,_RPoss),
			ord_union( LIds, RIds, PrnIds ),
			NewLeafDepth is LD - 1,
			NewLvs = [leaf(NewLeafDepth,PrnIds,PrnPoss)-ConnctVar|RrmLvs],
			CartOut = ec(PrunedCart,NewLvs,NewNdPsts).
1/3 :: edit( CartIn, CartOut ) :- 					% change
			CartIn = ec(Cart,Lvs,NdPsts),
			pmember( NdPsts, ChgNdPst, _RmNdPsts ),
			navigate( Cart, ChgNdPst, ChgNode, ChgCart, ConnctVar ),
			ChgNode = cart(F,Ch,Poss,L,R),
			type_of_modification( ChgType ),
			change_tree( ChgCart, ChgType ),

1/2 :: type_of_modification( variable ).
1/2 :: type_of_modification( rule ).
							

navigate( cart(_F,_Ch,_Poss,L,R), Psts, PruneNode, PrunedCart, ConnctVar ) :-

leaves_list_var_select( [H-HVar|T], Var, Leaf, RemLvsList ) :-
	HVar == Var,
	!,
	Leaf = H,
	RemLvsList = T.
leaves_list_var_select( [H|T], Var, Leaf, RemLvsList ) :-
	RemLvsList = [H|TRem],
	leaves_list_var_select( T, Var, Leaf, TRem ).

prunable_places( Leaf, _AccPath, Psts, Psts ) :-
	var( Leaf ), !.
prunable_places( cart(_F,_Ch,_Poss,L,R), AccPath, Psts, TailPsts ) :-
	var( L ), var( R ), 
	!,
	reverse( AccPath, ThisNodePath ),
	Psts = [ThisNodePath|DPsts],
	prunable_places( L, [l|AccPath], DPsts, LPsts ),
	prunable_places( R, [r|AccPath], LPsts, TailPsts ).
prunable_places( cart(_F,_Ch,_Poss,L,R), AccPath, Psts, TailPsts ) :-
	prunable_places( L, [l|AccPath], Psts, LPsts ),
	prunable_places( R, [r|AccPath], LPsts, TailPsts ).

% Cart = cart(Feature, ChoiceOnFeature, Possibles, LeftChild, RightChild ).

% cart( Feature, ChoiceOnFeature, LeftChild, RightChild )
% leaf( Depth, LeafDataIds, Possibles ).
%
split( leaf(0,Ids,P), ABeta, Cart ) :-
	pmember( P, f(FId,ChFeature), RmP ),
	pmember( ChFeature, s(ChSplit,ChL,ChR), RmChF  ),
	Rest = [f(FId,RmChF)|RmP],
	branch( Rest, ChL, PL ),
	branch( Rest, ChR, PR ),
	LvPrs = [leaf(1,ChL,PL)-L,leaf(1,ChR,PR)-R],
	Cart = cart(FId,ChSplit,L,R),
	split_cart( LvPrs, ABeta ).

split_cart( [H|T], Alpha/Beta ) :-
	!,
	pmember( [H|T], ChLfPr, RmLvPrs ),
	ChLfPr = leaf(D,Ids,P)-LV,
	Split is Alpha * exp((1+D),-Beta),
	[Split] : split_leaf( D, Ids, P, LV, RmLvPrs, NxLvPrs ),
	% [Split] : split_leaf( ChLfPr, ABeta, RmLvPrs, NxLvPrs ),
	split_cart( NxLvPrs, Alpha/Beta ).
split_cart( [], _ABeta ).

% Split: Split ::
split_leaf( D, Ids, [Hp|Tps], LV, LvPrs, NwLvPrs ) :-
			D1 is D + 1,
			pmember( [Hp|Tps], f(FId,ChFeature), RmP ),
			pmember( ChFeature, s(ChSplit,ChL,ChR), RmChF  ),
			Rest = [f(FId,RmChF)|RmP],
			branch( Rest, ChL, PL ),
			branch( Rest, ChR, PR ),
			LV = cart( FId, ChSplit, L, R ),
			NwLvPrs = [leaf(D1,ChL,PL)-L,leaf(D1,ChR,PR)-R|LvPrs].
% (1 - Split): Split :: split_leaf( D, Ids, P, leaf(D,Ids,P), LvPrs, LvPrs ).

:- pvars( pmember(L,_E,_R), [T-length(L,T)] ). 
1/X : X :: pmember( [H|T], H, T ).
(1 - 1/X) : X :: pmember( [H|T], El, [H|R] ) :-
				[(X - 1)]: pmember( T, El, R ).

branch( [], _Restr, [] ).
branch( [f(F,Psbls)|T], Restr, FRestr ) :-
	branch_possibles( Psbls, Restr, PrvRestr ),
	restricts( PrvRestr, F, FRestr, TFRestr ),
	% ( PrvRestr == [] ->
		% FRestr = TFRestr
		% ;
		% FRestr = [f(F,PrvRestr)|TFRestr]
	% ),
	branch( T, Restr, TFRestr ).

restricts( [], _F, TRestr, TRestr ).
restricts( [H|T], F, [f(F,[H|T])|TFRestr], TFRestr ).

branch_possibles( [], _Restr, [] ).
branch_possibles( [s(F,L,R)|T], Restr, Psbls ) :-
	ord_intersection( Restr, L, NwL ),
	ord_intersection( Restr, R, NwR ),
	not_vacuous_constructs( NwL, NwR, F, Psbls, TPsbls ),
	% ( (not_vacuous( NwL ), not_vacuous( NwR ) ) ->
		% Psbls = [s(F,NwL,NwR)|TPsbls]
		% ;
		% Psbls = TPsbls
	% ),
	branch_possibles( T, Restr, TPsbls ).

not_vacuous_constructs( [L1,L2,L3,L4,L5|LT], [R1,R2,R3,R4,R5|RT], F, All, Tail ) :-
	!,
	All = [s(F,[L1,L2,L3,L4,L5|LT],[R1,R2,R3,R4,R5|RT])|Tail].
not_vacuous_constructs( _, _, _F, All, All ).

% not_vacuous( [_,_|_] ).
:- module( edit, [variable_has_position_in_cart/3,edcart_to_concrete_cart/2] ).

variable_has_position_in_cart( Cart, Var, Position ) :-
	var( Cart ),
	!,
	Cart == Var,
	Position = []. % else, fail = is not here 
variable_has_position_in_cart( cart(_,_,_,L,R), Var, Position ) :-
	( variable_has_position_in_cart(L,Var,TPosition) -> 
		Position = [l|TPosition],
		;
		variable_has_position_in_cart( R, Var, TPosition ),
		!,
		Position = [r|TPosition]
	).

edcart_to_concrete_cart( ec(Cart,Lvs,_Psts), CartOut ) :-
	instantiate_leaves( Lvs ),
	five_arg_cart_to_four_arg( Cart, CartOut ).

instantiate_leaves( [] ).
instantiate_leaves( [H-H|T] ) :-
	instantiate_leaves( T ).

five_arg_cart_to_four_arg( cart(Ft,Ch,_Poss,L,R), cart(Ft,Ch,FrL,FrR) ) :-
	!,
	five_arg_cart_to_four_arg( L, FrL ),
	five_arg_cart_to_four_arg( R, FrR ).

navigate( cart(F,Ch,Poss,L,R), Psts, PruneNode, PrunedCart, ConnctVar ) :-
	( Psts = [H|T] -> 
		( H = l ->
			navigate( L, T, PruneNode, PrndL, ConnctVar ),
			PrunedCart = cart(F,Ch,Poss,PrndL,R)
			;
			navigate( R, T, PruneNode, PrndR, ConnctVar ),
			PrunedCart = cart(F,Ch,Poss,L,PrndR)
		)
		;
		PruneNode = cart(F,Ch,Poss,L,R),
		PrunedCart = ConnctVar
	).
