:- ensure_loaded( edit_mod ). 			% variable_has_position_in_cart/3,edcart_to_concrete_cart/2.
									% feature_choice_splits_ids/3,navigate/7.
:- ensure_loaded( library(lists) ).		% select/3, reverse/2.
:- ensure_loaded( library(ordsets) ).		% ord_union/3, ord_intersection/3,4.

% EdCart = Editable Cart = ec(Cart,LeavesList,NodePositions).
% LeavesList = [leaf(Depth,Ids,Possibles)-LeafV|T]
% NodePositions = [[],[{l,r}*]*] l-left, r-right. a list for each internal node should be present
% Cart = cart(Feature, ChoiceOnFeature, Possibles, LeftChild, RightChild ).
%		if Left/RightChild is a leaf, then it will be a variable, same as corresponding LeafV in LeavesList
%         this representation adds Possibles to node levels when compared to growing slp.
% Leaf = leaf(Depth,Ids,Possibles).
%
9/10 :: move( CartIn, CartOut ) :- edit( CartIn, CartMid ), move( CartMid, CartOut ).
1/10 :: move( CartIn, CartOut)  :- edcart_to_concrete_cart( CartIn, CartOut ).

1/3 :: edit( CartIn, CartOut ) :-  				% grow
			CartIn = ec(Cart,Lvs,NdPsts),
			pmember( Lvs, Leaf, RmLvs ),
			Leaf = leaf(D,Ids,Poss)-CartVar,
			variable_has_position_in_cart( Cart, CartVar, NwNodePos ),
			split_leaf( D, Ids, Poss, CartVar, RmLvs, NwLvs ),
			CartOut = ec(Cart,NwLvs,[NwNodePos|NdPsts] ).
1/3 :: edit( CartIn, CartOut ) :- 					% prune
			CartIn = ec(Cart,Lvs,NdPsts),
			prunable_places( Cart, [], PrnblPsts, [] ),
			pmember( PrnblPsts, PrunePst, _RemPrnbls ),
			navigate( Cart, PrunePst, 0, PruneNode, _PrnDpth, PrunedCart, ConnctVar ),
			PruneNode = cart(_PrnF,_PrnCs,PrnPoss,LVar,RVar),
			leaves_list_var_select( Lvs, LVar, LLv, LrmLvs ),
			leaves_list_var_select( LrmLvs, RVar, RLv, RrmLvs ),
			select( PrnPoss, NdPsts, NewNdPsts ),
			% LLv = leaf(VanishDepth,_,_),
			LLv = leaf(LD,LIds,_LPoss),
			RLv = leaf(_RD,RIds,_RPoss),
			ord_union( LIds, RIds, PrnIds ),
			NewLeafDepth is LD - 1,
			NewLvs = [leaf(NewLeafDepth,PrnIds,PrnPoss)-ConnctVar|RrmLvs],
			CartOut = ec(PrunedCart,NewLvs,NewNdPsts).
1/3 :: edit( CartIn, CartOut ) :- 					% change : i and ii
			CartIn = ec(Cart,Lvs,NdPsts),
			pmember( NdPsts, ChgNdPst, _RmNdPsts ),
			navigate( Cart, ChgNdPst, 0, ToChgNode, DChg, NwCart, ConnctVar ),
			ToChgNode = cart(F,Ch,Poss,L,R),
			feature_choice( F, Poss, NwF, FChoices, RmPoss ),
			pmember( FChoices,  s(NwSplit,ChL,ChR), RmChF  ).
			Rest = [f(FChoices,RmChF)|RmPoss],
			branch( Rest, LIds, PL ),
			branch( Rest, RIds, PR ),
			% 	90% sure that LIds and RIds are not vacuous, because split wouldnt ne in possibles
			NxD is DChg + 1,
			propagate_node_change( L, LIds, PL, NxD, Lvs, MidLvs, NwL ),
			propagate_node_change( R, RIds, PR, NxD, MidLvs, NwLvs, NwR ),
			ConnctVar = cart(NwF,NwSplit,Rest,NwL,NwR), 
				% IN ABOVE IS Rest the CORRECT ARGUMENT ?
			% var_cart_to_lvs_list_and_nd_psts( NwCart, NwLvs, NwNdPsts ),
			CartOut = ec(NwCart,NwLvs,NwNdPsts).


split_leaf( D, _Ids, [Hp|Tps], LV, LvPrs, NwLvPrs ) :-
	D1 is D + 1,
	pmember( [Hp|Tps], f(FId,ChFeature), RmP ),
	pmember( ChFeature, s(ChSplit,ChL,ChR), RmChF  ),
	Rest = [f(FId,RmChF)|RmP],
	branch( Rest, ChL, PL ),
	branch( Rest, ChR, PR ),
	LV = cart( FId, ChSplit, L, R ),
	NwLvPrs = [leaf(D1,ChL,PL)-L,leaf(D1,ChR,PR)-R|LvPrs].


1/2 :: feature_choice( F, Poss, FId, ChFeature, RmP ) :-
			pmember( Poss, f(FId,ChFeature), RmP ).
1/2 :: feature_choice( F, Poss, F, ChFeature, RmP ) :-
			select( f(F,ChFeature), Poss, RmP ).

propagate_node_change( Leaf, Ids, UpdtPoss, DChg, InLvs, OutLvs, Leaf ) :-
	var( Leaf ),
	!,
	leaves_list_var_select( InLvs, Leaf, GroundLeaf, RmLvs ),
	OutLvs = [leaf(DChg,Ids,UpdtPoss)|RmLvs].
propagate_node_change( Cart, UpdtPoss, Ids, DChg, InLvs, OutLvs, NwCart ) :-
	Cart = cart(F,Ch,_OldPoss,OldL,OldR),
	feature_choice_splits_ids( F, Ch, AllLeftIds ),
	ord_intersection( AllLeftIds, Ids, LIds, RIds ),
	updated_ids_propagate( LIds, RIds, F, Ch, DChg, OldL, OldR, UpdtPoss, Ids, InLvs, OutLvs, NwCart ).

updated_ids_propagate( LIds, RIds, F, Ch, D, OldL, OldR, UpdtPoss, Ids, InLvs, OutLvs, NwCart ) :-
	not_vacuous(LIds), not_vacuous( RIds ),
	!,
	select( f(F,UpdtChF), UpdtPoss, RmUpdtPoss ),
	select( s(Ch,UpdtChL,UpdtChR), UpdtChF, RmUpdtChF ),
	UpdtRest = [f(FChoices,RmUpdtChF)|RmUpdtPoss],
	branch( UpdtRest, LIds, UpdtLPoss ),
	branch( UpdtRest, RIds, UpdtRPoss ),
	NwCart = cart(F,CH,UpdtPoss,NwL,NwR),
	NxD is D + 1,
	propagate_node_change( OldL, UpdtLPoss, LIds, NxD, InLvs, MidLvs, NwL ),
	propagate_node_change( OldR, UpdtRPoss, RIds, NxD, MidLvs, OutLvs, NwR ).
updated_ids_propagate( _LIds, _RIds, _F, _Ch, D, _OldL, _OldR, UpdtPoss, Ids, InLvs, OutLvs, NwCart ) :-
	sub_cart_removes_from_leaves_list( Cart, InLvs, TOutLvs ),
	OutLvs = [leaf(D,Ids,UpdtPoss)-NwCart|TOutLvs].

leaves_list_var_select( [H-HVar|T], Var, Leaf, RemLvsList ) :-
	HVar == Var,
	!,
	Leaf = H,
	RemLvsList = T.
leaves_list_var_select( [H|T], Var, Leaf, RemLvsList ) :-
	RemLvsList = [H|TRem],
	leaves_list_var_select( T, Var, Leaf, TRem ).

prunable_places( Leaf, _AccPath, Psts, Psts ) :-
	var( Leaf ), !.
prunable_places( cart(_F,_Ch,_Poss,L,R), AccPath, Psts, TailPsts ) :-
	var( L ), var( R ), 
	!,
	reverse( AccPath, ThisNodePath ),
	Psts = [ThisNodePath|DPsts],
	prunable_places( L, [l|AccPath], DPsts, LPsts ),
	prunable_places( R, [r|AccPath], LPsts, TailPsts ).
prunable_places( cart(_F,_Ch,_Poss,L,R), AccPath, Psts, TailPsts ) :-
	prunable_places( L, [l|AccPath], Psts, LPsts ),
	prunable_places( R, [r|AccPath], LPsts, TailPsts ).

sub_cart_removes_from_leaves_list( Leaf, InLvs, OutLvs ) :-
	var( Leaf ), !,
	leaves_list_var_select( InLvs, Lef, _, OutLvs ).
sub_cart_removes_from_leaves_list( cart(_F,_Ch,_Poss,L,R), InLvs, OutLvs ) :-
	sub_cart_removes_from_leaves_list( L, InLvs, MidLvs ),
	sub_cart_removes_from_leaves_list( R, MidLvs, OutLvs ).

:- pvars( pmember(L,_E,_R), [T-length(L,T)] ). 
1/X : X :: pmember( [H|T], H, T ).
(1 - 1/X) : X :: pmember( [H|T], El, [H|R] ) :-
				[(X - 1)]: pmember( T, El, R ).

branch( [], _Restr, [] ).
branch( [f(F,Psbls)|T], Restr, FRestr ) :-
	branch_possibles( Psbls, Restr, PrvRestr ),
	restricts( PrvRestr, F, FRestr, TFRestr ),
	branch( T, Restr, TFRestr ).

branch_possibles( [], _Restr, [] ).
branch_possibles( [s(F,L,R)|T], Restr, Psbls ) :-
	ord_intersection( Restr, L, NwL ),
	ord_intersection( Restr, R, NwR ),
	not_vacuous_constructs( NwL, NwR, F, Psbls, TPsbls ),
	branch_possibles( T, Restr, TPsbls ).

restricts( [], _F, TRestr, TRestr ).
restricts( [H|T], F, [f(F,[H|T])|TFRestr], TFRestr ).

not_vacuous_constructs( [L1,L2,L3,L4,L5|LT], [R1,R2,R3,R4,R5|RT], F, All, Tail ) :-
	!,
	All = [s(F,[L1,L2,L3,L4,L5|LT],[R1,R2,R3,R4,R5|RT])|Tail].
not_vacuous_constructs( _, _, _F, All, All ).

not_vacuous( [_,_,_,_,_|_] ).
