% assumes logamma/2, and not_vacuous_length/1.

:- ensure_loaded( library(ordsets) ).		% list_to_ord_set/2.
:- ensure_loaded( library(system) ).		% file_exists/1.

:- elm( 'carts/src/kategories_to_ids' ).
:- ensure_loaded( library(kv_decompose) ).	% /3.
:- requires( werr/2 ).
:- requires( flatoms/2 ).

% Data = [[FeatureVec]-Kat|...]
data_to_feature_pairs( Data, Dataset, Splits, Ids, Lim ) :-
	kv_decompose( Data, FeatureVecs, Kategories ),
	kategories_to_ids( Kategories, 1, SerialIds ),
     length( SerialIds, IdsLength ),
     not_vacuous_length( Lim ),
     idsd_fprs_lim_load_file( Dataset, Lim, File ),
     Ids = ids(SerialIds,IdsLength),
     ( file_exists(File) ->
          ensure_loaded( File ),
          werr( bbs_file(File), info )
          ;
          open( File, write, Out ),
          % Ids = ids([1-IdsLength],IdsLength),
	     list_to_ord_set( Kategories, Kats ),
	     length( Kats, K ),
	     bb_put( all_categories, Kats ),
          portray_clause( Out, (:- bb_put(all_categories,Kats) ) ), 
	     bb_put( num_of_categories, K ),
          portray_clause( Out, (:- bb_put(num_of_categories,K) ) ), 
	     logamma( K, LogK ),
	     bb_put( knorm, LogK ), 
          portray_clause( Out, (:- bb_put(knorm,LogK) ) ), 
	     feature_bb_puts( FeatureVecs, 1, Out, Splits ),
          kv_decompose( Splits, ValidFs, _ ),
          length( ValidFs, NofValidFs ),
          bb_put( valid_features, ValidFs/NofValidFs ),
          portray_clause( Out, (:- bb_put(valid_features,ValidFs/NofValidFs) ) ),
          close( Out )
     ).
	
idsd_fprs_lim_load_file( Dataset, Lim, File ) :-
     number_codes( Lim, LimCs ),
     atom_codes( LimAtm, LimCs ),
     flatoms( [Dataset, '_nf', LimAtm, '.pl'], File ).

feature_bb_puts( [], I, Out, [] ) :-
     bb_put( nof_features, I ),
     portray_clause( Out, (:- bb_put(nof_features,I) ) ).
% feature_bb_puts( [H|T], I, [I-SplitVs|Tsps] ) :-
feature_bb_puts( [H|T], I, Out, Splits ) :-
     % write( in(I) ), nl,
	feature_val_skim( [H|T], 1, Fvs, IndexPrs, Rem ),
	sort( Fvs, UnqSrtFVals ),
	% write( srt( UnqSrtFVals ) ), nl,
	feature_index( UnqSrtFVals, Fvs, 1, SplitVs, IthIndex ),
	% open( indices, append, IOut ),
	% write( IOut, I+IthIndex ), nl( IOut ),
	% close( IOut ),
	NxI is I + 1,
     ( IthIndex == [] ->
          Splits = Tsps,
          werr( skipping_empty_splits(I,UnqSrtFVals), warning )
          % werr( [skipping_empty_splits(I,UnqSrtFVals)], '! ', '', user )
          ;
	     % bb_put( I, IthIndex ),
	     bb_put( I, IndexPrs ),
          portray_clause( Out, (:- bb_put(I,IndexPrs) ) ),
          ( (I mod 50 =:= 0) -> werr( [done(I)], '! ', '', user );
               true ),
          Splits = [I-SplitVs|Tsps]
     ),
	feature_bb_puts( Rem, NxI, Out, Tsps ).

feature_val_skim( [], _I, [], [], [] ).
feature_val_skim( [[H|R]|T], I, [H|TVs], [I-H|TPrs], Rem ) :-
	( R == [] -> TRem = Rem ; Rem = [R|TRem] ),
     NxI is I + 1,
	feature_val_skim( T, NxI, TVs, TPrs,TRem ).

feature_index( [], _Fvs, _Ith, [], [] ).
feature_index( [H|T], Fvs, I, SplitVs, Index ) :-
	( number(H) ->
		arith_dichotomy_ord_int( Fvs, H, 1, 0, -1, 0, 0, LessOrEq, Greater, LgL, LgG )
		;
		val_dichotomy( Fvs, H, 1, LessOrEq, Greater )
	),
     not_vacuous_length( Lim ),
	( (LgL > Lim, LgG > Lim) ->
		SplitVs = [H|TSplitVs],

		Index = [H-LessOrEq|TIndex]
		;
		TSplitVs = SplitVs, TIndex = Index
	),
	NxI is I + 1,
	feature_index( T, Fvs, NxI, TSplitVs, TIndex ).

arith_dichotomy_ord_int( [], _, _I, Fst, Lst, AcL, LgG, LOrE, [], LgL, LgG ) :-
     ( Fst =:= 0 -> 
          LOrE = [],
          LgL is AcL
          ;
          FstPlus1 is Fst + 1,
          compare( Op, FstPlus1, Lst ),
          op_adds_to_lore_list( Op, Fst, Lst, [], AddLg, LOrE ),
          LgL is AcL + AddLg
     ).
arith_dichotomy_ord_int( [H|T], Pivot, I, Fst, Lst, AcL, AcG, LOrE, G, LgL, LgG ) :-
	% This is a hack, data shouldnt have na_ (not available) entries
	% since the likelihood does nt deal with them.
	( H==na_ -> 
		( Pivot @< H -> 
               broken_lore( Fst, Lst, NxtFst, NxtLst, LOrE, AddL, TLOrE ),
               NxL is AcL + AddL, % TLOrE = LOrE,
               G = [I|TG], NxG is AcG + 1
			;
               add_to_lore( Fst, Lst, I, NxtFst, NxtLst, LOrE, AddL, TLOrE ),
			% LOrE = [I|TLOrE], 
               G = TG
		)
		;
	     ( Pivot < H ->
               broken_lore( Fst, Lst, NxtFst, NxtLst, LOrE, AddL, TLOrE ),
	     	% TLOrE = LOrE,
               NxL is AcL + AddL,
               G = [I|TG], NxG is AcG + 1
	     	;
	     	% LOrE = [I|TLOrE],
               add_to_lore( Fst, Lst, I, NxtFst, NxtLst, LOrE, AddL, TLOrE ),
               NxL is AcL + AddL,
               G = TG, NxG is AcG
	     )
	),
	NxI is I + 1,
	arith_dichotomy_ord_int( T, Pivot, NxI, NxtFst, NxtLst, NxL, NxG, TLOrE, TG, LgL, LgG ).

% add_to_lore/7 and broken_lore/6 deal with adding 
% elements or ranges to the less or equal list.
% 
add_to_lore( 0, _Lst, I, NxtFst, NxtLst, LOrE, 0, TLOrE ) :-
     !,
     NxtFst is I, 
     NxtLst is I,
     LOrE = TLOrE.
add_to_lore( Fst, Lst, I, NxtFst, NxtLst, LOrE, Add, TLOrE ) :-
     ( (Lst + 1 =:= I) -> 
          NxtFst is Fst,
          LOrE = TLOrE, Add is 0
          ;
          NxtFst is I,
          ( Fst =:= Lst ->
               LOrE = [Fst|TLOrE], Add is 1
               ;
               LOrE = [Fst-Lst|TLOrE], Add is Lst - Fst + 1
          )
     ),
     NxtLst is I.

broken_lore( 0, _Lst, 0, -1, LOrE, 0, LOrE ) :-
     !.
broken_lore( Fst, Lst, 0, -1, LOrE, Add, TLOrE ) :-
     FstPlus1 is Fst + 1,
     compare( Op, FstPlus1, Lst ),
     op_adds_to_lore_list( Op, Fst, Lst, TLOrE, Add, LOrE ).

% op_adds_to_lore_list( Op, Fst, Lst, Tail, [Fst|Tail] ).
% Op is the result of comparing Fst + 1 to Lst.
%
op_adds_to_lore_list( >, Fst, _Lst, Tail, 1, [Fst|Tail] ).
op_adds_to_lore_list( =, Fst, Lst, Tail, 2, [Fst,Lst|Tail] ).
op_adds_to_lore_list( <, Fst, Lst, Tail, Add, [Fst-Lst|Tail] ) :-
     Add is Lst - Fst + 1.

arith_dichotomy( [], _, _I, [], [] ).
arith_dichotomy( [H|T], Pivot, I, LOrE, G ) :-
	% This is a hack, data shouldnt have na_ (not available) entries
	% since the likelihood does nt deal with them.
	( H==na_ -> 
		( Pivot @< H -> 
			TLOrE = LOrE, G = [I|TG]
			;
			LOrE = [I|TLOrE], G = TG
		)
		;
	( Pivot < H ->
		TLOrE = LOrE, G = [I|TG]
		;
		LOrE = [I|TLOrE], G = TG
	)
	),
	NxI is I + 1,
	arith_dichotomy( T, Pivot, NxI, TLOrE, TG ).

val_dichotomy( [], _, _I, [], [] ).
val_dichotomy( [H|T], Pivot, I, LOrE, G ) :-
	( Pivot @< H ->
		TLOrE = LOrE, G = [I|TG]
		;
		LOrE = [I|TLOrE], G = TG
	),
	NxI is I + 1,
	val_dichotomy( T, Pivot, NxI, TLOrE, TG ).
