%
:- requires( not_vacuous/1 ).

:- ensure_loaded( library(ordsets) ).		% ord_add_element/3.
									% list_to_ord_set/2.
:- ensure_loaded( library(lists) ).		% member/2.

:- requires( kv_decompose/3 ).
:- ensure_loaded( kategories_to_ids ).
:- elm( 'bns/src/llhood/logamma' ).	% /2.

% J = number of features.
% T = number of data points.
% We assume all data points feature all features.
% Data = [[f_1:1,f_1:2,...,f_1:j]-C1,[f_2:1,f_2:2,...,f_2:J]-C2,
%           ...,[f_T:1,f_T:2,...,f_T:j]].
% Feature types: list of J l or n identifiers for qualitative, and quantative
data_to_cart_root( Data, FTypes, Ids, Psbls ) :-
	kv_decompose( Data, FeatureVecs, Kategories ),
	kategories_to_ids( Kategories, 1, Ids ),
	list_to_ord_set( Kategories, Kats ),
	length( Kats, K ),
	bb_put( all_categories, Kats ),
	bb_put( num_of_categories, K ),
	% sum_of_logs_of_k_ones( K, 0, SlgK ),
	logamma( K, LogK ),
	bb_put( knorm, LogK ),
	feature_types_to_possibles( FTypes, 1, FeatureVecs, Psbls ).

feature_types_to_possibles( [], _F, _EmptyData, [] ).
feature_types_to_possibles( [H|T], F, Data, [f(F,HPsbls)|TPsbls] ) :-
	feature_domain( Data, [], Domain, FeatureData, ChpdData ),
	( H == l ->
		qualitative_domain_to_possibles( Domain, FeatureData, HPsbls )
		;
		quantitative_domain_to_possibles( Domain, FeatureData, HPsbls )
	),
	NxF is F + 1,
	feature_types_to_possibles( T, NxF, ChpdData, TPsbls ).

feature_domain( [], Dmn, Dmn, [], [] ).
feature_domain( [[Hf|TFs]|T], Acc, Dmn, [Hf|MrFs], [TFs|TCData] ) :-
	ord_add_element( Acc, Hf, NxAcc ),
	feature_domain( T, NxAcc, Dmn, MrFs, TCData ).

qualitative_domain_to_possibles( Domain, FeatureData, Psbls ) :-
	ord_to_subsets( Domain, [], Subsets ),
	qualitative_subsets_to_possibles( Subsets, FeatureData, Psbls ).

qualitative_subsets_to_possibles( [], _FeatureData, [] ).
qualitative_subsets_to_possibles( [H|T], FeatureData, Psbls ) :-
	qualitative_set_splits( FeatureData, H, 1, Left, Right ),
	( (not_vacuous(Left),not_vacuous(Right)) ->
		Psbls = [s(H,Left,Right)|TPsbls]
		;
		Psbls = TPsbls
	),
	qualitative_subsets_to_possibles( T, FeatureData, TPsbls ).

qualitative_set_splits( [], _Set, _N, [], [] ).
qualitative_set_splits( [H|T], Set, N, Left, Right ) :-
	NxN is N + 1,
	( ord_member( H, Set ) -> 
		Left = [N|TLeft],
		Right = TRight
		;
		Left = TLeft,
		Right = [N|TRight]
	),
	qualitative_set_splits( T, Set, NxN, TLeft, TRight ).

quantitative_domain_to_possibles( [], _FeatureData, [] ).
quantitative_domain_to_possibles( [H|T], FeatureData, Psbls ) :-
	quantitative_value_splits( FeatureData, H, 1, Left, Right ),
	( (not_vacuous(Left),not_vacuous(Right)) ->
		Psbls = [s(H,Left,Right)|TPsbls]
		;
		Psbls = TPsbls
	),
	quantitative_domain_to_possibles( T, FeatureData, TPsbls ).

quantitative_value_splits( [], _Val, _N, [], [] ).
quantitative_value_splits( [H|T], Val, N, Left, Right ) :-
	NxN is N + 1,
	( H =< Val -> 
		Left = [N|TLeft],
		Right = TRight
		;
		Left = TLeft,
		Right = [N|TRight]
	),
	quantitative_value_splits( T, Val, NxN, TLeft, TRight ).

% not_vacuous( [_,_,_,_,_|_] ). % assume this is in memory

% I think this is ok. 
% from Hogg and Tannis p77. we need Unordered sapmling
% without replacement n! / r!(n-r!)  with n = nof objects, r = nof selected.
% then summate over r... for n = 4 , 1 + 4 + 6 + 4 = 15
% | ?- ord_to_subsets( [a,b,c,d], Subs ), length( Subs, Length ).
% Length = 15 ? 
%
ord_to_subsets( [], Subsets, Subsets ).
ord_to_subsets( [H|T], AccSets, Subsets ) :-
	findall( SetH, (member(Set,AccSets),
				 ord_add_element(Set,H,SetH) ), NewSets ),
	append( AccSets, [[H]|NewSets], NxAccSets ),
	ord_to_subsets( T, NxAccSets, Subsets ).
