
:- ensure_loaded( '../../src/init_lib' ).

:- ensure_loaded( library(read_terms) ).			% /2.
:- ensure_loaded( library(sys_info) ).				% /1.
:- ensure_loaded( library(matching_terms_in_list) ).	% /5.
:- ensure_loaded( library(lists) ).	% memberchk/2, append/3.

% expected(excl(20,'%'),2,1.0).
% means 20 % from the start of the chain has been removed.
% That is the 1st and 2nd chain points in this case.
% 1.0 is the reduced chain's loss average.
%

add_exp_exclude_pcent( Opts ) :-
	matching_terms_in_list( pcent(Pcent), Pcent, Opts, Pcents, Files ),
	add_exp_exlude_pcent_from_files( Files, Pcents ).

add_exp_exlude_pcent_from_files( [], _ ).
add_exp_exlude_pcent_from_files( [H|T], Pcents ) :-
	( atom_concat(Stem,'.gz',H) -> WasGz=true ; Stem=H, WasGz=false ),
	atom_codes( Stem, StemCs ),
	append( BaseCs, [0'.,0'l,A,B], StemCs ),
	append( BaseCs, [0'.,0'e,A,B], ExpHCs ),
	atom_codes( ExpH, ExpHCs ),
	read_terms( ExpH, ExpTerms ),
	file_pcentages( Pcents, ExpTerms, H, WasGz, ToGz, FPcents ),
	( FPcents == [] -> 
			true
			;
			num_of_lines_in_file( Stem, Lns ),
			pcentages_to_points( FPcents, Lns, UnsPoints ),
			open( Stem, read, InStm ),
			read( InStm, RdTrm ),
			keysort( UnsPoints, Points ),
			expected_points( RdTrm, InStm, 1, Points, 0, PointSums, Sum ),
			open( ExpH, append, App ),
			append_expectation_of_sum_points_to( PointSums, App, Lns, Sum ),
			close( App )
	),
	( ToGz == true -> 
		atom_concat( 'gzip ', Stem, Gzip ),
		sys_info( Gzip )
		;
		true
	),
	add_exp_exlude_pcent_from_files( T, Pcents ).

file_pcentages( [], _ExpT, _H, _WGz, _, [] ).
file_pcentages( [Hpc|TPCs], ExpTerms, H, WGz, TGz, FPCs ) :-
	( memberchk(expected(excl(Hpc,'%'),_Lgt,Val),ExpTerms) ->
		write(already_exists(f(H),p(Hpc),v(Val))), nl
		;
		FPCs = [Hpc|TFPCs],
		( WGz == true ->
			atom_concat( 'gunzip ', H, Gunzip ),
			sys_info( Gunzip ),
			NGz = false,
			TGz = true
			;
			NGz = false
		)
	),
	file_pcentages( TPCs, ExpTerms, H, NGz, TGz, TFPCs ).

expected_points( end_of_file, _In, _I, Points, Sum, [], Sum ) :-
	!,
	( Points \== [] ->
		write( user_error, non_empty_points(Points) ),
		nl( user_error )
		;
		true
	).
expected_points( l(IVal), In, I, Points, AccSum, PSums, Sum ) :-
	NxSum is AccSum + IVal,
	( Points = [I-Pcent|TPoints] ->
		PSums = [I-Pcent-NxSum|TPSums]
		;
		TPoints = Points,
		TPSums  = PSums
	),
	read( In, NxTrm ),
	NxI is I + 1,
	expected_points( NxTrm, In, NxI, TPoints, NxSum, TPSums, Sum ).

append_expectation_of_sum_points_to( [], _App, _Lgt, _Sum ).
append_expectation_of_sum_points_to( [N-Pcent-SumToN|T], App, Lgt, Sum ) :-
	ExpExcToN is (Sum - SumToN) / (Lgt - N),
	writeq( App, expected(excl(Pcent,'%'),N,ExpExcToN) ),
	write( App, '.' ), nl( App ),
	append_expectation_of_sum_points_to( T, App, Lgt, Sum ).

num_of_lines_in_file( File, Lns ) :-
	open( File, read, In ),
	read( In, T ),
	num_of_terms_in_file_1( T, In, 0, Lns ),
	close( In ).

num_of_terms_in_file_1( end_of_file, _In, Lns, Lns ) :-
	!.
num_of_terms_in_file_1( _T, In, Acc, Lns ) :-
	Nxt is Acc + 1,
	read( In, NxT ),
	num_of_terms_in_file_1( NxT, In, Nxt, Lns ).

pcentages_to_points( [], _Lns, [] ).
pcentages_to_points( [H|T], Lns, Points ) :-
	( H > 100 ->
		write( user_error, error_in_percentage(H) ), nl(user_error),
		Points = TPoints
		; 
		HPnt is integer(((H * Lns) / 100)),
		Points = [HPnt-H|TPoints]
	),
	pcentages_to_points( T, Lns, TPoints ).
