
:- ensure_loaded( library(lists) ).		% memberchk/2.
:- ensure_loaded( '../../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( select_compound_w_gunzip/3 ).
:- requires( gzip_files/3 ).

loss_files_to_exp_loss_points( Opts ) :-
	opts_have_help( Opts ),
	!, nl,
	write( 'loss_files_to_exp_loss_points:    a script for producing expected loss step' ), nl, 
	write( '                                  files from loss files.' ), nl, 
	write( '                                  Its output files can be fed into'), nl,
	write( '                                  mcmcms/auxil/gsplay.pl .' ), nl,
	write( ' Options.' ), nl,
	write( '       compress/1     whether to use gzip on the output. Def. is true.' ), nl,
	write( '                      and at least one should be given.' ), nl,
	write( '       out/1          returns output filenames.' ), nl,
	write( '       step/1         every how many input losses should an expected' ), nl,
	write( '                      loss should be recorded. Default is 1.' ), nl,
	write( '                      Non compound options are taken to be input loss files.' ), nl, nl, nl.



loss_files_to_exp_loss_points( Opts ) :-
	( memberchk(step(Step),Opts) -> 
		true
		;
		write( using_default_step(1) ), nl,
		Step = 1
	),
	% findall( F, member(lfile(F),Opts), NestFs ),
	select_compound_w_gunzip( Opts, _, Fs, Gunzipped ),
	% ( Fs = [] ->
		% write( user_error, 'Didnt find any lfile/1 options, quiting.'),
		% nl( user_error )
		% ;
		( memberchk(compress(Bgz),Opts) -> true ; Bgz = true ),
		loss_files_to_exp_loss_point_files( Fs, Step, Bgz, OutFs ),
	% ),
	gzip_files( Gunzipped ),
	( memberchk(out(OutFs),Opts) -> true ; true ).

loss_files_to_exp_loss_point_files( [], _Step, _Bgz, [] ).
loss_files_to_exp_loss_point_files( [F|Fs], Step, Bgz, [OutH|OutT] ) :-
	( atom_concat( Stem, '.gz', F ) -> 
		atom_concat( 'gunzip ', F, Gunzip ),
		system( Gunzip ),
		loss_file_to_exp_loss_point_file( Stem, Step, PntsFile ),
		atom_concat( 'gzip ', Stem, Gzip ),
		system( Gzip )
		;
		loss_file_to_exp_loss_point_file( F, Step, PntsFile )
	),
	( Bgz == true -> 
		atom_concat( 'gzip ', PntsFile, GzipPoints ), 
		system( GzipPoints ),
		atom_concat( PntsFile, '.gz', OutH )
		;
		OutH = PntsFile
	),
	loss_files_to_exp_loss_point_files( Fs, Step, Bgz, OutT ).

loss_file_to_exp_loss_point_file( File, Step, Gile ) :-
	number_codes( Step, StepCs ),
	% atom_codes( Ext, [0'_,0'e|StepCs] ),
	atom_codes( File, FileCs ),
	append( StemCs, [0'.,0'l,A,B], FileCs ),
	append( StemCs, [0'.,0'g,A,B,0'_,0's|StepCs], GileCs ),
	atom_codes( Gile, GileCs ),
	open( File, read, In ),
	open( Gile, write, Out ),
	read_term( In, T, [] ),
	loss_stream_to_exp_loss_point_stream( T, 1, Step, 0, 0, In, Out ),
	close( In ),
	close( Out ).

loss_stream_to_exp_loss_point_stream( end_of_file, I, _S, Sum, It, _In, Out ) :-
	!,
	( I =:= 1 ->
		true
		;
		EPoint is Sum / It,
		write( Out, EPoint ), nl( Out )
	).
	% write( user_error, sum(Sum) ), nl( user_error ),
	% write( user_error, lgt(It ) ), nl( user_error ).
loss_stream_to_exp_loss_point_stream( l(L), I, I, Sum, It, In, Out ) :-
	!,
	NxSum is Sum + L, 
	NxIt is It + 1,
	EPoint is NxSum / NxIt,
	write( Out, EPoint ), nl( Out ),
	read_term( In, T, [] ),
	loss_stream_to_exp_loss_point_stream( T, 1, I, NxSum, NxIt, In, Out ).
loss_stream_to_exp_loss_point_stream( l(L), I, St, Sum, It, In, Out ) :-
	NxSum is Sum + L, 
	NxIt is It + 1,
	NxI is I + 1,
	read_term( In, T, [] ),
	loss_stream_to_exp_loss_point_stream( T, NxI, St, NxSum, NxIt, In, Out ).
