:- ensure_loaded(library(lists)).
:- ensure_loaded(library(ordsets)).
:- ensure_loaded( cpt  ).        % for cpts
:- ensure_loaded( reachable  ).        % d-separation test

%bn([f1-[xx(2),xy(2)],m1-[xx(1),xy(1)],xx(1)-[xx(4),xy(3)],xx(2)-[],xx(3)-[],xx(4)-[],xy(1)-[xx(3),xy(2)],xy(2)-[],xy(3)-[]]).

% needs cpt/3 like this
% cpt('A'-a,['B'-b,'C'-c],0.4).

% and 

% some way of getting all possible values for a set of variables

/*---------------------------------------------------------------------
  Predicate: joint_prob(+Inst,+BN,-Prob)
       Args: Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..] (ordered set)
	     BN a Bayes net as a list of families
             Prob: = P(Inst) according to BN
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/
joint_prob([Var-Val],_BN,Prob) :-  % if only one instantiated then easy
	!,
	cpt(Var-Val,[],Prob).      % just the marginal

joint_prob(Inst,BN,Prob) :-
	unextract(Inst,InstVars),
	descendents(BN,Descendents),
	reduce_bn(BN,BN,InstVars,Descendents,RedBN),
	normalise(RedBN,RedBN,NormBN),
	descendents(NormBN,NormDescendents),
	write(joint_prob(Inst,NormBN,NormDescendents,Prob)),nl,
	joint_prob(Inst,NormBN,NormDescendents,Prob).


normalise([],_BN,[]).
normalise([Child-Parents|T],BN,[Child-NewParents|T2]) :-
	 findall(P,
		 (
		   member(P,Parents),
		   member(P-_,BN)
		 ),
		 NewParents
		),
	 normalise(T,BN,T2).

%reduce_desc([],_RedBN,[]).
%reduce_desc([Var-Desc|T],RedBN,[Var-Desc|T2]) :-
%	memberchk(Var-_Parents,RedBN),
%	!,
%	reduce_desc(T,RedBN,T2).
%reduce_desc([_|T],RedBN,T2) :-
%	reduce_desc(T,RedBN,T2).

reduce_bn([],_BN,_InstVars,_Descendents,[]).
reduce_bn([Child-Parents|T],BN,InstVars,Descendents,[Child-Parents|T2]) :-
	memberchk(Child,InstVars),  % can't get rid of instantiated ones!
	!,
	reduce_bn(T,BN,InstVars,Descendents,T2).
reduce_bn([Child-Parents|T],BN,InstVars,Descendents,[Child-Parents|T2]) :-
	once((memberchk(Child-Desc,Descendents),
	member(D,Desc),
	memberchk(D,InstVars),
	ancestors(Child,BN,Ancestors),
	ord_add_element(Ancestors,Child,Ancestors1))),
	member(A,Ancestors1),	                 % an ancestor (or self)
	(
	  memberchk(A,InstVars)                  % which is instantiated or ...
	;
	  member(A-AsDs,Descendents),              
	  member(AsD,AsDs),                      % with a descendent
	  AsD \= D,                              % other than D
	  memberchk(AsD,InstVars)                % which is instantiated
	),
	!,
	reduce_bn(T,BN,InstVars,Descendents,T2).
reduce_bn([_Family|T],BN,InstVars,Descendents,T2) :-          % can be deleted
	reduce_bn(T,BN,InstVars,Descendents,T2).

/*---------------------------------------------------------------------
  Predicate: joint_prob(+Inst,+BN,+Descendants,-Prob)
       Args: Inst: An instantiation of variables in BN of the form
	     [var3-val13,var4-val42..] (ordered set)
	     BN a Bayes net as a list of families
	     Descendents list with elts Var-[VD1,VD2,..]
             Prob: = P(Inst) according to BN
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/

% empty product
joint_prob([],_BN,_Desc,1.0).

% Single variable
% Special case: in pedigrees all variables have the same marginal!
% just a short cut
joint_prob([Var-Val],_BN,_Desc,Prob) :-
	!,
	cpt(Var-Val,[],Prob).


joint_prob(Inst,BN,Descendents,Prob) :-
	member(Var-_Val,Inst),
	member(Var-Desc,Descendents),
	\+ ((member(X-_,Inst),member(X,Desc))), % none of Var's desecdents in Inst
	!,
	memberchk(Var-Parents,BN),           % find parents
	(
	  extract(Parents,Inst,ParentInst)   % succeds if parents are there
	->
	  split_off(Var,ParentInst,Inst,BN,Descendents,Prob)
	;
	  reachable(BN,Var,[],Reachable),    % find vars dependent on Var
	  (
	    (member(Y,Reachable),Y\=Var,memberchk(Y-_,Inst))  % damn, depends on an elt of Inst
	  ->
	    sum_out(Var,Parents,Inst,BN,Descendents,Prob)
	  ;
	    split_off(Var,[],Inst,BN,Descendents,Prob)
	  )
	).



sum_out(Var,Parents,Inst,BN,Descendents,Prob) :-
	get_given(Parents,Inst,MiniInst,Others),
	memberchk(Var-Val,Inst),
	Val=Marker-_,		           % sneakily grab marker, 
	possible_vals(Marker,PossVals),    % in cpt.pl
	findall(
		ParentInst,
		(
		  poss_inst(Others,PossVals,Inst,OtherInst),
		  append(MiniInst,OtherInst,ParentInst)
		),
		ParentsInsts
	       ),
	sum_out1(ParentsInsts,Var-Val,Inst,BN,Descendents,0,Prob).

get_given([],_Inst,[],[]).
get_given([H|T],Inst,[H-Val|T1],Others) :-
	memberchk(H-Val,Inst),
	!,
	get_given(T,Inst,T1,Others).
get_given([H|T],Inst,MiniInst,[H|T2]) :-
	get_given(T,Inst,MiniInst,T2).


poss_inst([],_PossVals,_Inst,[]).
poss_inst([H|T],PossVals,Inst,[H-Val|T2]) :-
	member(Val,PossVals),
	poss_inst(T,PossVals,Inst,T2).
		
sum_out1([],_VarVal,_Inst,_BN,_Descendents,Prob,Prob).
sum_out1([H|T],VarVal,Inst,BN,Descendents,In,Out) :-
	cpt(VarVal,H,P1),
	ord_union(H,Inst,NewInst),
	joint_prob(NewInst,BN,Descendents,P2),
	Mid = In + (P1*P2),
	!,
	sum_out1(T,VarVal,Inst,BN,Descendents,Mid,Out).


split_off(Var,ParentInst,Inst,BN,Descendents,Prob) :-
	select(Var-Val,Inst,Rest),
	cpt(Var-Val,ParentInst,P1),
	!,
	joint_prob(Rest,BN,Descendents,P2),
	Prob is P1 * P2.
	

extract([],_Inst,[]).
extract([Var|Vars],Inst,[Var-Val|Rest]) :-
	memberchk(Var-Val,Inst),
	extract(Vars,Inst,Rest).
	 

unextract([],[]).
unextract([Var-_Val|Vars],[Var|Rest]) :-
	unextract(Vars,Rest).	




descendents(BN,Descendents) :-
	children(BN,BNChildren),               % Children is BN upside-down
	findall(
		X-Descendents,
		(
		  member(X-_,BNChildren),
		  descendents(X,BNChildren,Descendents)
		),
		BNDesc
	       ),
	list_to_ord_set(BNDesc,Descendents).


% get children from a BN
children(BN,BNChildren) :-
	findall(
		Parent-Children,
		(member(Parent-_,BN),children(Parent,Children,BN)),
		BNKids
	       ),
	list_to_ord_set(BNKids,BNChildren).

% get all children of a given node
children(Parent,Children,BN) :-
	findall(Child,child(Parent,Child,BN),Kids),
	list_to_ord_set(Kids,Children).

% when is Parent a parent of Child in BN
child(Parent,Child,BN) :-
	member(Child-Parents,BN),
	member(Parent,Parents).

descendents(X,BNChildren,XsDescendents) :-
	findall(
		XsD,
		descendent(X,XsD,BNChildren),
		XsDs
	       ),
	list_to_ord_set(XsDs,XsDescendents).

ancestors(X,BN,XsAncestors) :-
	findall(
		XsA,
		ancestor(X,XsA,BN),
		XsAs
	       ),
	list_to_ord_set(XsAs,XsAncestors).


% Y is a descendent of X
descendent(X,Y,BNChildren) :-
	memberchk(X-Kids,BNChildren),
	(
	  member(Y,Kids)
	;
	  member(Z,Kids),
	  descendent(Z,Y,BNChildren)
	).

ancestor(X,Y,BN):-
	memberchk(X-Parents,BN),
	(
	  member(Y,Parents)
	;
	  member(Z,Parents),
	  ancestor(Z,Y,BN)
	).


























%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% NOT USED

%% If just given the parents, then done
%% Parents is just names
%% Given has values attached to names
%joint_cond_prob([Var-Val],Given,BN,Prob) :-
%	memberchk(Var-Parents,BN),
%	match(Given,Parents),
%	!,
%	cpt(Var-Val,Given,Prob).


%% try to split into two conditionally indepndent parts
%joint_cond_prob(Inst,Given,BN,Prob) :-
%	partition(Inst,Inst1,Inst2),
%	ci(Inst1,Inst2,Given,BN),
%	!,
%	joint_cond_prob(Inst1,Given,BN,Prob1),
%	joint_cond_prob(Inst2,Given,BN,Prob2),
%	Prob is Prob1*Prob2.

%% P(A=a,B=b,C=c|D=d) = P(A=a,B=b|C=c,D=d) * P(C=c|D=d)
%joint_cond_prob(Inst,Given,BN,Prob) :-
%	select(Var-Val,Inst,InstLess),         %just grab first one
%	!,
%	ord_add_element(Given,Var-Val,GivenMore),
%	joint_cond_prob(InstLess,GivenMore,Prob1),
%	joint_cond_prob([Var-Val],Given,Prob2),
%	Prob is Prob1*Prob2.


%partition([],[],[]).
%partition([H|T],[H|T2],L3) :-
%	partition(T,T2,L3).
%partition([H|T],L2,[H|T3]) :-
%	partition(T,L2,T3).

%ci(Inst1,Inst2,Given,BN) :-
%	\+ cd(Inst1,Inst2,Given,BN).

%% true if active path between X \in Inst1 and Y \in Inst2
%% with Given as Evidence
%cd(Inst1,Inst2,Given,BN) :-
%	member(X-_,Inst1),
%	reachable(BN,X,Given,Reachable),  % in reachable.pl
%	member(Y-_,Inst2),
%	member(Y,Reachable).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% UTILITY
	
%match([],[]).
%match([H-_|T],Vars) :-
%	select(H,Vars,Rest),
%	!,
%	match(T,Rest).

%%get_desc([],_,[]).
%%get_desc([A-AsKids|T],Children,[A-AsDesc|T2]):-
%%	findall(X,descendent(AsKids,X,Children),AsD),
%%	list_to_ord_set(AsD,AsDesc),
%%	get_desc(T,Children,T2)

%%descendent(AsKids,B,_Children) :-
%%	  member(B,AsKids)
%%descendent(AsKids,B,Children) :-
%%	  member(C,AsKids),
%%	  member(C-CsKids,Children),
%%	  descendent(CsKids,B,Children)

%%children([],Out,Out).
%%children([Child-Parents|T],In,Out) :-
%%	add_kids(Parents,Child,In,Mid),
%%	children(T,Mid,Out).

%%add_kids([],_Child,Out,Out).
%%add_kids([H|T],Child,In,Out) :-
%%	add_child(H,Child,In,Mid),
%%	add_kids(T,Child,Mid,Out).

%%add_child(H,Child,In,Out) :-
%%	select(H-Kids,In,Rest),
%%	!,
%%	ord_add_element(Kids,Child,NewKids),
%%	ord_add_element(Rest,H-NewKids,Out).
%%add_child(H,Child,In,Out) :-
%%	ord_add_element(In,H-[Child],Out).


