% run/1 uses the run_data/n, residing in file RunsFile, to run a number
% of experiments described in the RunsFile.
% see doc/users_manual.ps for details.

% system libraries
:- ensure_loaded( library(system) ).		
		% file_exists/1.
:- ensure_loaded( library(lists) ).		
		% memberchk/2, reverse/2, select/3, append/3.

:- ensure_loaded( 'src/init_lib' ).
:- ensure_loaded( library(flatoms) ).				% /2.
:- ensure_loaded( library(sys_info) ).				% /1.
:- ensure_loaded( library(read_nth_unifying_term) ).	% /4.
:- ensure_loaded( library(is_a_directory) ).			% /1.
:- ensure_loaded( library(dirpath_will_exist) ).		% /1.
:- ensure_loaded( library(werr) ).					% /1.
:- ensure_loaded( library(dbg) ).					% dbg_these/1.
:- ensure_loaded( library(number_abbreviation) ).		% /2.
:- ensure_loaded( library(delete_trailing) ).		% /3.
:- ensure_loaded( library(replace_many_with_one_code) ).%/4.
:- ensure_loaded( mcmcms ).		

:- multifile( exec_extension/2 ).
:- dynamic( exec_extension/2 ).

:- ensure_loaded( library(esr) ).				% /1.
% :- ensure_loaded( library(slp_file_location) ).	% /3.
:- ensure_loaded( library(non_dir_exists) ).	% /1.

:- pl( yap(_Y), true, ensure_loaded(library(number_atom)) ). % /2.

% :- bb_put( sr, off ). % To print on screen. also known in mcmcms.

/* 
run( [] ).
equivalent to
run( [runs_file('runs/simple'),results_dir(test)] ).
the following is the simplest test possible.
run( [id(1),runs_file('runs/tiny'),results_dir(test)] ).
run( [id(1),runs_file('runs/seeds_effect_lnai_8pu'),results_dir(test)] ).
Following for displaying bns.
run( [display(next_display_at), id(2),runs_file('runs/small'),results_dir(test)] ).
As above, but with second iterations models being identical :
run( [display(next_display_at), id(1),runs_file('runs/small'),results_dir(test)] ).
run( [display(every_single_one_of_them),id(1)] ). obsolete ?
*/

run :- run( [] ).

run( PrvArgs ) :-
	bb_put( run_bnames, [] ),
	clean_up_run_options( PrvArgs, InArgs ),
	( select( display(PredName), InArgs, Args ) ->
		DispCall =.. [PredName,Iter,Jump],
		current_output( Out ),
		bb_put( display_at, (Iter,DispCall,wait_for_return_or_integer(Out,Iter,Jump),Out,Jump) ),
		Display = true
		;
		Display = false,
		Args = InArgs
	),
	( memberchk(print(These),Args) ->
		true ; These =  [28] ),
	% bb_put( dbg_these, These ),
	dbg_these( These ),
	( memberchk(runs_file(RunsFile),Args) -> true ; RunsFile = 'rruns' ),
	( memberchk(source_dir(Source),Args) -> true ; Source = '.' ),
	locate_file( RunsFile, '.pl', runs, Source, fail, RunsSourceFile ),
	( memberchk(results_dir(ResDir),Args) -> true
		; ResDir = test
		),
	absolute_file_name( ResDir, AbsResDir ),
	( memberchk(initial(InitialM),Args) ->
		bb_put( initial, InitialM )
		;
		true
	),
	( is_a_directory(AbsResDir) -> true ;
		dirpath_will_exist( AbsResDir )
	),
	%% flatoms( [ResDir,'/std_out_n_err'], Err ),
	%% open( Err, append, Estream ),
	%% current_output( OldOutStream ),
	%% bb_put( old_out_stream, OldOutStream ),
	%% prolog_flag( user_error, OldErrStream ), % do this without esr/1
	%% bb_put( old_err_stream, OldErrStream ),
	%% bb_put( err_stream, Estream ),
	% consult( RunsSourceFile ),
	open_null_stream( Null ),
	with_error_to_stream( compile(RunsSourceFile), Null ), 
	( memberchk(id(_),Args) ->
		findall( Id, member(id(Id),Args), Ids)
		; 
		Ids = 1 
	),
	( bb_get(before_experiments,BECall)-> call(BECall); true ),
	run_1( Ids, Args, Source, Display, ResDir, RunsSourceFile ).

run_1( Ids, Args, Source, Display, ResDir, RnsFl ) :-
	( is_list(Ids) -> Ids = [Id|RmIds] ; Id is Ids, RmIds is Ids + 1 ),
	run_data( Id, K, Slp, MS, Data, SGl, OutF, Rs, Hot, RndS, Exts ),
	!,
	hot_loads( Hot ),
	set_up_bp_and_ms( K, MS, Id, KName, Pat, P ),
	%% esr( set_output( Estream ) ),
	%% esr( set_prolog_flag( user_error, Estream ) ),
	number_atom( RndS, Sat ),
	Data =.. [DAtm|_DArgs],
	run_filename( KName, Pat, MS, DAtm, OutF, Rs, Sat, Filename ),
	bb_get( run_bnames, OldEnames ),
	bb_put( run_bnames, [Filename|OldEnames] ),
	% ( memberchk(basename(Filename),Args) -> true ; true ),
	seed_id_to_seed_atoms( RndS, S1, S2, S3 ),
	flatoms( [ResDir,'/',Filename], DirF ),
	flatoms( [DirF,'.stats'], Sfile ),
	absolute_file_name( Sfile, SfileLoc ),
	open( SfileLoc, write, Stats ),
	( bb_get(last_data,ld(LastData,LstDf)) ->
		LastData =.. [LstDAtm|LstArgs],
		( LstDAtm == DAtm ->
			DataFile = LstDf
			;
			atom_concat( clean_up_, LstDAtm, CleanDAtm ),
			CleanUp =.. [CleanDAtm|LstArgs],
			call( CleanUp ),
			locate_file( DAtm, '.pl', data, Source, true, DataFile ),
			( DataFile == null -> 
				% with_error_to_stream( reconsult(DataFile), Stats )
				pl(swi(_),consult(DataFile),user:reconsult(DataFile))
				;
				( DataFile == null -> 
					true
					; 
					werr( [['Warning: non existing data file, for: ',DataFile]] )
				)
			)
		)
		;
		locate_file( DAtm, '.pl', data, Source, true, DataFile ),
		( file_exists(DataFile) -> % with_error_to_stream( reconsult(DataFile), Stats )
			pl(swi(_),consult(DataFile),user:reconsult(DataFile))
			; ( DataFile == null -> true
				; werr( [['Warning: non existing data file, for: ',DataFile]] )) )
	),
	close( Stats ),
	locate_file( Slp, '.slp', slps, Source, fail, TheSlp ),
	% flatoms( [DirF,'.stats'], Sfile ),
     nl, nl, nl,
	werr( running_experiment(Id), info ), nl,
     werr( stats_file(SfileLoc), info ),
	( memberchk(progress(BAtm,Perc),Args) -> 
		calc_percentiles( Perc, Perc, Rs, [HPtl|TPtiles] ),
		bb_put( progress, pts(BAtm,HPtl,TPtiles) )
		; true ),
	( Display == true -> bb_put( next_display_iteration, 1 ) ; true ),
	% ( Id =:= 2 -> spy(preds_to_file/2) ; true  ),
	bb_put( last_data, ld(Data,DataFile) ),
     ( memberchk(model_type(ModTp),Args) -> true; ModTp = bn ),
     absolute_file_name( DirF, DirFLoc ),
	mcmcms( DirFLoc, SfileLoc, KName, ModTp, TheSlp, P, Rs, Data, Hot, S1, S2, S3, SGl ),
	garbage_collect,
	garbage_collect_atoms,
	esr( do_extensions( Exts, DirFLoc ) ),
     ( memberchk(gzip(false), Args) ->
          true
          ;
	     flatoms( ['gzip -f ',DirFLoc], Gzip ),
	     esr( sys_info( Gzip ) )
     ),
	bb_put( files, files(ResDir,RnsFl,TheSlp,DataFile) ),
	!,
	run_1( RmIds, Args, Source, Display, ResDir, RnsFl ).
run_1( _Ids, Args, _Src, _Disp, _ResDir, _RnsFl ) :-
	( memberchk(copy(Copy),Args)-> true; Copy = [runfile] ),
	nl( user_error ),
	copy_files( Copy ),
	( memberchk(basenames(Bnames),Args) ->
		bb_get( run_bnames, Bnames )
		;
		true
	).

logical_prolog_flag( A, B ) :-
	atomic( A ),
	prolog_flag( A, B ),
	!.

do_extensions( Unfolds, DirF ) :-
	( (var(Unfolds);Unfolds==all) ->
		write( user_error, 'Extensions: ' ),
		Goal = (exec_extension(Ext,DirF),write(user_error,Ext),write(user_error,' ')),
		findall( Ext, Goal, _Exts ),
		write( user_error, '; end of extensions.' ), nl( user_error )
		;
		( is_list(Unfolds) ->
			write( user_error, 'Successful extensions: ' ),
			exec_extensions(Unfolds,DirF),
			write( user_error, '; end of extensions' ), nl( user_error )
			; 
			( Unfolds == none ->
				werr( [['No extensions.']] )
				;
				werr( [[wrong_extensions(Unfolds)]] ), 
				halt
			)
		)
	).

exec_extensions( [], _DirF ).
exec_extensions( [H|T], DirF ) :-
	( once( exec_extension( H, DirF ) ) ->
		write( user_error, H ), write( user_error, ' ' )
		;
		true
	),
	exec_extensions( T, DirF ).

seed_id_to_seed_atoms( Rnd, S1, S2, S3 ) :-
	bb_get( mcmcms_top_dir, RunDir ),
	atom_concat( RunDir,'/auxil/seeds.pl', SeedsFile ), 
	read_nth_unifying_term( SeedsFile, 1, seed(Rnd,_), Seed ),
	Seed = seed( _, rand(X1,X2,X3) ),
	number_atom( X1, S1 ),
	number_atom( X2, S2 ),
	number_atom( X3, S3 ).

set_up_bp_and_ms( K, MS, Id, KName, Pat, P ) :-
	K =.. [KName|PssPArg],
	( initial_bp_str(KName,_,_) -> true
	  ; Mess = [['Unknown backtracking strategy \'',KName,'\'.'],
	            ['Skipping experiment \'',Id,'\'.'],[]
	           ],
		werr( Mess ),
		fail
	),
	( PssPArg = [] ->
		Pat = ''
		;
		PssPArg = [P],
		( number( P ) ->
			number_codes( P, PCs ) 
			;	
			P = PprK-PprV, 
			number_codes( PprK, PprKCs ),
			number_codes( PprV, PprVCs ),
			append( PprKCs, [0'_|PprVCs], PCs )
		),
		replace_many_with_one_code( PCs, [0'.], 0'_, PRplCs ),
		atom_codes( Pat, PRplCs )
	),
	( (MS==rm;MS==fm)-> bb_put( msd, MS )
	  ; Mess = [['Unknown model space definition \'',MS,'\'.'],
	            ['Skipping experiment \'',Id,'\'.'],[]
	           ],
		werr( Mess ),
		fail
	).

run_filename( NameK, Pat, MS, DAtm, ExC, Repeats, Sat, Filename ) :-
% run_filename( Kernel, P, Ord, Cnsts, Repeats, Data, S, Filename ) :-
	number_abbreviation( Repeats, Abrv ),
	% Kernel = [KName, Pat],
	flatoms( ['tr_',NameK,Pat,'_',MS,'_',DAtm,'_',ExC,'_i',Abrv,'_','_s',Sat], Filename ).

%
wait_for_return( Out ) :-
	write( Out, ' press return to continue ' ), get_code( _C ), nl(Out).

every_single_one_of_them( _ ).

next_display_at( Iter, Jump ) :-
	% bb_get( last_display_iteration, Last ),
	bb_get( next_display_iteration, Next ),
	( Next = jump(N) ->
		( Jump == jump -> 
			( N > 1 ->
				NxN is N - 1,
				bb_put( next_display_iteration, jump(NxN) ),
				fail
				;
				Jump == jump
			)
		)
		;
		Iter >= Next
	).

wait_for_return_or_integer( Out, Iter, _Jump ) :-
	write( Out, ' press return or give next iteration to display ([+]N jumps; 0 halts; -N skips) : ' ),
	get_code( C ), % nl(Out),
	( (C >= 0'0, C =< 0'9) -> 
		get_integer_from_get_codes( [C], Integer ),
		NxDisplay is Iter + Integer
		;
		( (C =:= 0'-,get_integer_from_get_codes([],Integer)) -> 
			write( Out, 'skipping to ' ), write_n( Out, Integer ),
			write( Out, ' jump ...' ), nl( Out ),
			NxDisplay = jump(Integer)
			;	
			Integer is 1,
			NxDisplay is Iter + Integer,
			consume_codes_to_return( C )
		)
	),
	!,
	% write( Out, nx_disp_at(NxDisplay) ), nl( Out ),
	bb_put( next_display_iteration, NxDisplay ).

get_integer_from_get_codes( Acc, Integer ) :-
	get_code( C ),
	% write( user_error, last_code(C) ), nl( user_error ),
	( (C >= 0'0, C =< 0'9) -> 
		get_integer_from_get_codes( [C|Acc], Integer )
		;
		reverse( Acc, IntCs ),
		number_codes( Integer, IntCs ),
		( Integer =:= 0 -> halt ; true )
		% consume_codes_to_return( C )
	).
	% write( user_error, out_of ), nl( user_error ).

consume_codes_to_return( 10 ) :- !.
consume_codes_to_return( _ ) :-
	get_code( C ),
	consume_codes_to_return( C ).

locate_file( File, Ext, Dir, Source, True, FileLoc ) :-
	(working_directory(Old,Source) ; (Source\== '.',Old='.')),
	( non_dir_exists(File) ->
	  absolute_file_name( File, FileLoc )
	  ;
	  atom_concat( File, Ext, FileExt ),
	  ( non_dir_exists(FileExt) ->
	    absolute_file_name( FileExt, FileLoc )
	    ;
	    atom_concat( Dir, '/', Trail ),
	    atom_concat( Trail, File, TrFile ),
	    ( non_dir_exists(TrFile) ->
	      absolute_file_name( TrFile, FileLoc )
		 ;
		 atom_concat( Trail, FileExt, TrFileExt ),
		 ( non_dir_exists(TrFileExt) ->
	        absolute_file_name( TrFileExt, FileLoc )
		    ;
		    working_directory(_,Old),
		    FileLoc = File,
		    call( True )
		 )
	    )
	  )
	),
	!,
	working_directory( _, Old ).
locate_file( File, Ext, Dir, Source, _, _Loc ) :-
	werr( [['Could not locate file ', File, '[', Ext,']', ' in sub-directory ', Dir, ' of ', Source]] ),
	abort.

copy_files( all ) :-
	!,
	copy_files( [script(rrun),runfile,slp,data] ).
copy_files( none ) :- !.
copy_files( [] ) :- !.
copy_files( [H|T] ) :-
	bb_get( files, files(ResDir,Runs,Slp,Data) ),
	( H = script(InScript) ->
		locate_file( InScript, '.pl', '.', '.', fail, LocScript ),
		absolute_file_name(LocScript,AbsScript),
		( non_dir_exists(AbsScript) ->
			( atom_concat(ResDir,_,AbsScript) ->
				true % no need to copy, write message
				;
				copy_file_to( AbsScript, ResDir )
			)
			;
			werr( [['copy_files/1: cannot locate scipt ',InScript,' for copying; skippping']] )
		)
		;
		( H == runfile -> File = Runs 
		  ; ( H == slp -> File = Slp 
		      ; ( H == data -> File = Data 
			     ; werr( [['Unrecognised selector: ',H,' in copy/1 option of run/1, skipping.']] )
				) )
		),
		copy_file_to( File, ResDir )
	),
	copy_files( T ).

copy_file_to( File, Dir ) :-
	( (var(File) ; atom_concat(Dir,_,File)) ->
			% maybe we should add ignore_message(File)
			true 
			;
			atom_concat( ' ', Dir, SpDir ),
			atom_concat( File, SpDir, FromTo ),
			atom_concat( 'cp ', FromTo, SysCopy ),
			sys_info( SysCopy )
	).

with_error_to_stream( Goal, Stream ) :-
	pl( swi(_P), 
			( stream_property( Err, alias(user_error) ),
			  set_stream( Stream, alias(user_error) ),
			  call( Goal ),
			  set_stream( Err, alias(user_error) )
			)
			,
			( prolog_flag( user_error, Old, Stream ),
			  call( Goal ),
			 prolog_flag( user_error, _TheS, Old ) )
		).

calc_percentiles( Perc, _Step, _Repeats, Ptiles ) :-
	Perc > 100, 
	!,
	Ptiles = [].
calc_percentiles( Perc, Step, Repeats, [F|M] ) :-
	F is integer( Perc * Repeats / 100 ),
	NxP is Perc + Step,
	calc_percentiles( NxP, Step, Repeats, M ).

write_n( Out, N ) :-
	write( Out, N ), 
	(N=:=1 -> Tl=st; (N=:=2 -> Tl=nd; (N=:=3 -> Tl=rd; Tl = th))),
	write( Out, Tl ).

clean_up_run_options( In, Out ) :-
	Reco = [display(_),print(_),runs_file(_),source_dir(_),model_type(_),
	        results_dir(_),progress(_,_),initial(_),basenames(_),gzip(_)],
	Multi = [id(_)],
	copy_term( Reco, Treco ),
	remove_variables_from_options( In, NoVar ),
	remove_multiple_options( NoVar, Multi, NoMulti, AllMulti ),
	clean_up_run_options( Reco, NoMulti, Clean, Rem ),
	append( AllMulti, Clean, Out ),
	print_options_warnings( Rem, Treco ).

remove_variables_from_options( [], [] ).
remove_variables_from_options( [H|T], NoVar ) :-
	( var(H) -> 
		werr( [['Removing variable: ',H,' from run/1 options.']] ),
		NoVar = TNoVar
		;
		NoVar = [H|TNoVar]
	),
	remove_variables_from_options( T, TNoVar ).

remove_multiple_options( [], _Multi, [], [] ).
remove_multiple_options( [H|T], Multi, NoMulti, All ) :-
	( \+ memberchk(H,Multi) ->
		NoMulti = [H|TNoMulti],
		All = Tall
		;
		NoMulti = TNoMulti,
		All = [H|Tall]
	),
	remove_multiple_options( T, Multi, TNoMulti, Tall ).
		
clean_up_run_options( [], Opts, [], Opts ).
clean_up_run_options( [H|T], Opts, Clean, Rem ) :-
	( select(H,Opts,NxOpts) ->
		Clean = [H|TClean]
		;
		NxOpts = Opts,
		Clean = TClean
	),
	clean_up_run_options( T, NxOpts, TClean, Rem ).

print_options_warnings( [], _Reco ).
print_options_warnings( [H|T], Reco ) :-
	( \+ memberchk(H,Reco) ->
		werr( [['Throwing away, run/1, unrecognised option: ',H,'.']] )
		;
		werr( [['Throwing away, run/1, multiply defined: ',H,'.']] )
	),
	print_options_warnings( T, Reco ).

hot_loads( [] ).
hot_loads( [_|_] ) :-
	bb_get( mcmcms_top_dir, Dir ),
	atom_concat( Dir, '/src/power_tempering.pl', ELoad ),
	ensure_loaded( ELoad ).
